from pySubnetSB.compatibility_collection import CompatibilityCollection  # type: ignore
from pySubnetSB.named_matrix import NamedMatrix   # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.reaction_constraint import ReactionConstraint  # type: ignore

import numpy as np
from scipy.special import factorial  # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False


class TestCompatibilityCollection(unittest.TestCase):

    def setUp(self) -> None:
        self.collection = CompatibilityCollection(2, 3)

    def testNumAssignment(self):
        if IGNORE_TEST:
            return
        for size in np.random.randint(2, 7, 100):
            collection = CompatibilityCollection(size, size)
            [collection.add(i-1, np.random.randint(0, size, size)) for i in range(1, size+1)]
            expanded_collection, _ = collection.expand()
            if expanded_collection.shape[0] > 0:
                log10_expanded_collection = max(0, np.log10(expanded_collection.shape[0]))
            else:
                log10_expanded_collection = 0
            self.assertTrue(np.abs(collection.log10_num_assignment - log10_expanded_collection) < 2)

    def testPrune(self):
        if IGNORE_TEST:
            return
        log10_max_assignment = 4.0
        for size in np.random.randint(5, 30, 100):
            collection = CompatibilityCollection(size, size)
            [collection.add(i-1, range(i)) for i in range(1, size+1)]
            new_collection, is_changed = collection.prune(log10_max_assignment)
            result = (collection.log10_num_assignment <= log10_max_assignment) and (not is_changed)
            result = result or (collection.log10_num_assignment > log10_max_assignment) and is_changed
            self.assertTrue(result)
            if not is_changed:
                self.assertEqual(new_collection, collection)
            else:
                self.assertNotEqual(new_collection, collection)

    def testExpand(self):
        if IGNORE_TEST:
            return
        size = 5
        collection = CompatibilityCollection(size, size)
        [collection.add(i-1, range(i)) for i in range(1, size+1)]
        arr, is_truncated = collection.expand()
        self.assertFalse(is_truncated)
        self.assertLessEqual(arr.shape[0], 1)
        self.assertEqual(arr.shape[1], size)

    def testExpandTruncated(self):
        if IGNORE_TEST:
            return
        size = 40
        collection = CompatibilityCollection(size, size)
        [collection.add(i-1, range(i)) for i in range(1, size+1)]
        arr, is_truncated = collection.expand(max_num_assignment=10)
        self.assertTrue(isinstance(arr, np.ndarray))
        self.assertTrue(is_truncated)

    def testExpandReductionInSize(self):
        if IGNORE_TEST:
            return
        fill_size = 5
        sizes = np.random.randint(5, 10, 20)
        for size in sizes:
            network = Network.makeRandomNetworkByReactionType(size, size)
            large_network = network.fill(num_fill_reaction=fill_size*size, num_fill_species=fill_size*size)
            large_constraint = ReactionConstraint(large_network.reactant_nmat, large_network.product_nmat)
            constraint = ReactionConstraint(network.reactant_nmat, network.product_nmat)
            compatibility_collection = constraint.makeCompatibilityCollection(
                  large_constraint).compatibility_collection
            arr, is_truncated = compatibility_collection.expand(max_num_assignment=1e10)
            for row in arr:
                for idx in range(len(row)):
                    expected_name = "J" + str(idx)
                    if expected_name in large_network.reaction_names:
                        break
                else:
                    self.assertTrue(False)

    def testBug(self):
        if IGNORE_TEST:
            return
        list_of_lists = [[3, 11, 16, 17], [2, 8, 14, 15, 17], [1], [2, 4, 5, 7, 12], [17],
                                    [2, 4, 5, 7, 12], [0, 8], [6, 13], [6, 13], [3, 11, 16, 17], [3, 11, 16, 17]]
        collection = CompatibilityCollection.makeFromListOfLists(list_of_lists)
        arr, _ = collection.expand()
        self.assertEqual(arr.shape[1], len(list_of_lists))


if __name__ == '__main__':
    unittest.main(failfast=True)


'''Constraint objects for subgraphs.'''

"""
Constraint objects have two kinds of constraints:
   1. Equality constraints that must have an exact match.
   2. Inequality constaints where the reference must have a value no greater than its target.

Constraints are NamedMatrices such that the columns are constraints and the rows are instances.
For reaction networks, instances are either species or reactions.

Subclasses are responsible for implementing:
  property: row_names - List[str] of row names
  property: description - str of description
  property: categorical_nmat - NamedMatrix of categorical constraints, those that are tested by equality
  property: numerical_enumerated_nmat - NamedMatrix of enumerated constraints
  property: bitwise_enumerated_nmat - NamedMatrix of boolean constraints such that the target
    must have a one bit for every bit in the reference.
  property: one_step_nmat - NamedMatrix of one step transitions
  property: to_reaction_arr - np.ndarray of reaction types for each reaction
"""

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.named_matrix import NamedMatrix, NULL_NMAT # type: ignore
from pySubnetSB.compatibility_collection import CompatibilityCollection # type: ignore

import json
import scipy.special  # type: ignore
import numpy as np
import scipy  # type: ignore
import string  # type: ignore
from typing import List, Tuple

NULL_INT = -1
ALL = "all"
NONE = "none"


#####################################
class CompatibilityCollectionResult(object):

    def __init__(self, compatibility_collection:CompatibilityCollection,
          equality_compatibility_nmat=NULL_NMAT,
          inequality_compatibility_numerical_nmat=NULL_NMAT,
          inequality_compatibility_bitwise_nmat=NULL_NMAT,
          reference_equality_nmat:NamedMatrix=NULL_NMAT,
          target_equality_nmat:NamedMatrix=NULL_NMAT,
          reference_numerical_inequality_nmat:NamedMatrix=NULL_NMAT,
          target_numerical_inequality_nmat:NamedMatrix=NULL_NMAT,
          reference_bitwise_inequality_nmat:NamedMatrix=NULL_NMAT,
          target_bitwise_inequality_nmat:NamedMatrix=NULL_NMAT):
        self.compatibility_collection = compatibility_collection
        self.equality_compatibility_nmat = equality_compatibility_nmat
        self.inequality_compatibility_numerical_nmat = inequality_compatibility_numerical_nmat
        self.inequality_compatibility_bitwise_nmat = inequality_compatibility_bitwise_nmat
        self.reference_equality_nmat = reference_equality_nmat
        self.target_equality_nmat = target_equality_nmat
        self.reference_numerical_inequality_nmat = reference_numerical_inequality_nmat
        self.target_numerical_inequality_nmat = target_numerical_inequality_nmat
        self.reference_bitwise_inequality_nmat = reference_bitwise_inequality_nmat
        self.target_bitwise_inequality_nmat = target_bitwise_inequality_nmat


#####################################
class ReactionClassification(object):
    MAX_REACTANT = 3
    MAX_PRODUCT = 3
    MAX_ENCODING = 1000
    LABELS = ["null", "uni", "bi", "multi"]
    NUM_CLASSIFICATION = len(LABELS)**2
    REACTION_CLASSES:List['ReactionClassification'] = []
    REACTION_BITS = [1 << i for i in range(NUM_CLASSIFICATION)]

    def __init__(self, num_reactant:int, num_product:int):
        self.num_reactant = min(num_reactant, self.MAX_REACTANT)
        self.num_product = min(num_product, self.MAX_PRODUCT)
        self.encoding = self.num_reactant*100 + self.num_product
        self.index = self.getIndex(self.num_reactant, self.num_product)

    @classmethod
    def getIndex(cls, num_reactant, num_product):
        return int(num_reactant*len(cls.LABELS) + num_product)

    def __repr__(self)->str:
        result = f"{self.LABELS[int(self.num_reactant)]}-{self.LABELS[int(self.num_product)]}"
        return result

    @classmethod 
    def getReactionClassifications(cls)->List['ReactionClassification']:
        """Gets a list of reaction classifications."""
        if len(cls.REACTION_CLASSES) > 0:
            return cls.REACTION_CLASSES
        pair_arr = np.array([f"{cls.LABELS[n]}-{cls.LABELS[m]}" for n in range(cls.MAX_REACTANT+1)
              for m in range(cls.MAX_PRODUCT+1)])
        idx_arr = np.array([cls.getIndex(n, m) for n in range(cls.MAX_REACTANT+1) for m in range(cls.MAX_PRODUCT+1)])
        cls.REACTION_CLASSES = list(pair_arr[idx_arr])
        return cls.REACTION_CLASSES

    @classmethod 
    def makeReactionClassificationMatrix(cls, reaction_names:List[str],
          classifications:List['ReactionClassification'])->NamedMatrix:
        """Makes a matrix whose rows are reaction names and columns are classifications.

        Args:
            reaction_names: List[str]
            classifications: List[ReactionClassification]

        Returns:
            NamedMatrix: rows are reactions; columns are reaction classifications
        """
        zero_arr = np.zeros(cls.NUM_CLASSIFICATION)
        arrays = []
        for classification in classifications:
            arr = zero_arr.copy()
            arr[classification.index] = 1
            arrays.append(arr)
        result = NamedMatrix(np.array(arrays), row_names=reaction_names, column_names=cls.REACTION_CLASSES)
        return result
    

#####################################
class Constraint(object):

    def __init__(self, reactant_nmat:NamedMatrix, product_nmat:NamedMatrix, is_subnet:bool=True):
        """
        Args:
            reactant_nmat: NamedMatrix
            product_nmat: NamedMatrix
            num_row: int
            num_column: int
        """
        self.reactant_nmat = reactant_nmat
        self.product_nmat = product_nmat
        self.is_subnet = is_subnet
        # Calculated
        self._reaction_classes = ReactionClassification.getReactionClassifications()
        self.num_species, self.num_reaction = self.reactant_nmat.num_row, self.reactant_nmat.num_column
        self.reaction_classification_arr = self.classifyReactions()
        # Outputs are categorical_nmat and enumerated_nmat, which are implemented by subclass
        self._is_subnet_initialized = False
        self._equality_nmat = NULL_NMAT
        self._numerical_inequality_nmat = NULL_NMAT
        self._bitwise_inequality_nmat = NULL_NMAT

    @property
    def num_row(self)->int:
        return self.one_step_nmat.num_row

    @classmethod
    def calculateLog10UnconstrainedPermutation(cls, reference_size:int, target_size:int)->float:
        """Calculates the log10 of the number of permutations to examine if no constraint is applied.

        Args:
            reference_size: int (number of species and reactions)
            target_size: int (number of species and reactions)

        Returns:
            float
        """
        def calculatePermutation(num_reference:int, num_target:int)->int:
            result = scipy.special.factorial(num_target) /  \
                  scipy.special.factorial(num_target - num_reference, exact=True)
            return result
        log_permutation = np.log10(calculatePermutation(reference_size, target_size))
        return 2*log_permutation
    
    ################ SUBCLASS MUST IMPLEMENT ################
        
    @property
    def description(self)->str:
        raise NotImplementedError("self be implemented by subclass.")
        
    @property
    def row_names(self)->List[str]:
        raise NotImplementedError("row_names be implemented by subclass.")
        
    @property
    def categorical_nmat(self)->NamedMatrix:
        raise NotImplementedError("numerical_categorical_nmat be implemented by subclass.")
    
    @property
    def numerical_enumerated_nmat(self)->NamedMatrix:
        raise NotImplementedError("numerical_enumerated_nmat be implemented by subclass.")

    @property
    def bitwise_enumerated_nmat(self)->NamedMatrix:
        raise NotImplementedError("logical_enumerated_nmat be implemented by subclass.")
    
    @property
    def one_step_nmat(self)->NamedMatrix:
        raise NotImplementedError("one_step_nmat be implemented by subclass.")
    
    @property
    def to_reaction_arr(self)->np.ndarray:
        raise NotImplementedError("to_reaction_arr be implemented by subclass.")

    ########################################################
    
    def _initializeSubset(self):
        # Initialize the equality and inequality NamedMatrices
        if self._is_subnet_initialized:
            return
        if self.is_subnet:
            self._equality_nmat = self.categorical_nmat
            self._numerical_inequality_nmat = self.numerical_enumerated_nmat
            self._bitwise_inequality_nmat = self.bitwise_enumerated_nmat
        else:
            if self.categorical_nmat == NULL_NMAT:
                self._equality_nmat = self.numerical_enumerated_nmat
            elif self.numerical_enumerated_nmat == NULL_NMAT:
                self._equality_nmat = self.categorical_nmat
            else:
                self._equality_nmat = NamedMatrix.hstack([self.categorical_nmat, self.numerical_enumerated_nmat])
            if self.bitwise_enumerated_nmat != NULL_NMAT:
                self._equality_nmat = NamedMatrix.hstack([self._equality_nmat, self.bitwise_enumerated_nmat])
            self._numerical_inequality_nmat = NULL_NMAT
        self._is_subnet_initialized = True
   
    @property
    def equality_nmat(self)->NamedMatrix:
        self._initializeSubset()
        return self._equality_nmat
    
    @property
    def numerical_inequality_nmat(self)->NamedMatrix:
        self._initializeSubset()
        return self._numerical_inequality_nmat

    @property
    def bitwise_inequality_nmat(self)->NamedMatrix:
        self._initializeSubset()
        return self._bitwise_inequality_nmat

    def __repr__(self)->str:
        return "Categorical\n" + str(self.categorical_nmat) + "\n\nEnumerated\n" +  str(self.numerical_enumerated_nmat)

    def __eq__(self, other)->bool:
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        if self.reactant_nmat != other.reactant_nmat:
            return False
        if self.product_nmat != other.product_nmat:
            return False
        return True
    
    def setSubset(self, is_subnet:bool)->None:
        self.is_subnet = is_subnet
        self._is_subnet_initialized = False

    def copy(self):
        return self.__class__(self.reactant_nmat.copy(), self.product_nmat.copy())
    
    def serialize(self)->str:
        """Serializes the boundary values."""
        return json.dumps({cn.S_ID: self.__class__.__name__,
                           cn.S_REACTANT_NMAT: self.reactant_nmat.serialize(),
                           cn.S_PRODUCT_NMAT: self.product_nmat.serialize(),
                           })

    @classmethod
    def deserialize(cls, string:str)->'Constraint':
        """Deserializes the boundary values."""
        dct = json.loads(string)
        if not cls.__name__ in dct[cn.S_ID]:
            raise ValueError(f"Expected {cls} but got {dct[cn.S_ID]}")
        reactant_nmat = NamedMatrix.deserialize(dct[cn.S_REACTANT_NMAT])
        product_nmat = NamedMatrix.deserialize(dct[cn.S_PRODUCT_NMAT])
        return cls(reactant_nmat, product_nmat)
    
    def classifyReactions(self)->List[ReactionClassification]:
        """Classify the reactions based on the number of reactants and products.

        Returns:
            List[ReactionClassification]
        """
        classifications = []
        for idx in range(self.num_reaction):
            num_reactant = np.sum(self.reactant_nmat.values[:, idx])
            num_product = np.sum(self.product_nmat.values[:, idx])
            classifications.append(ReactionClassification(num_reactant, num_product))
        return classifications

    @classmethod 
    def calculateBooleanCompatibilityVector(cls, reference_constraint_nmat:NamedMatrix,
              other_constraint_nmat:NamedMatrix, is_equality:bool=True,
              is_bitwise:bool=False)->Tuple[np.ndarray, NamedMatrix]:  # type: ignore
        """
        Calculates the compatibility of two matrices of constraints.

        Args:
            other: Constraint
            is_equality: bool (default: True) Equality or inequality
            is_bitwise: bool (default: False) Bitwise comparisons

        Returns:
            np.ndarray - vector of booleans
                index n represents i*reference_num_row + jth row in other
            NamedMatrix - matrix of booleans where the rows are the same as the input matrices
        """
        if reference_constraint_nmat.num_column != other_constraint_nmat.num_column:
            raise ValueError("Incompatible number of columns.")
        #
        reference_num_row = reference_constraint_nmat.num_row
        other_num_row = other_constraint_nmat.num_row
        # Check for a null
        num_column = other_constraint_nmat.num_column
        # Calculate the CompatibilityCollection
        #    Create the self array with repeated blocks of each row
        reference_arr = np.concatenate([reference_constraint_nmat.values]*other_num_row, axis=1)
        reference_arr = np.reshape(reference_arr, (reference_num_row*other_num_row, num_column))
        #    Create the other array with all values of each row
        other_arr = np.reshape(other_constraint_nmat.values, other_num_row*num_column)
        other_arr = np.concatenate([other_arr]*reference_num_row)
        other_arr = np.reshape(other_arr, (reference_num_row*other_num_row, num_column))
        # Calculate the compatibility boolean vector
        if is_equality:
            satisfy_arr = reference_arr == other_arr
        else:
            if is_bitwise:
                satisfy_arr = np.bitwise_and(reference_arr, other_arr) == reference_arr
            else:
                satisfy_arr = reference_arr <= other_arr
        satisfy_nmat = NamedMatrix(satisfy_arr,
                           column_names=reference_constraint_nmat.column_names)
        result_arr = np.sum(satisfy_arr, axis=1) == num_column
        return result_arr, satisfy_nmat
    
    def makeCompatibilityCollection(self, target:'Constraint',
          is_diagnostic:bool=False)->CompatibilityCollectionResult:
        """
        Makes a collection of compatible constraints.

        Args:
            target: Constraint
            is_diagnostic: bool (default: False) If True, return a diagnostic matrix

        Returns:
            CompatibilityCollectionResult

        """
        # Initializations
        equality_compatibility_nmat = NULL_NMAT
        inequality_compatibility_numerical_nmat = NULL_NMAT
        inequality_compatibility_bitwise_nmat = NULL_NMAT
        # Calculate the compatibility of the constraints
        if self.equality_nmat != NULL_NMAT:
            is_equality_compatibility = True
            equality_compatibility_arr, equality_compatibility_nmat =  \
                 self.calculateBooleanCompatibilityVector(self.equality_nmat, target.equality_nmat, is_equality=True)
        else:
            is_equality_compatibility = False
        if self.numerical_inequality_nmat != NULL_NMAT:
            is_inequality_compatibility = True
            inequality_compatibility_arr, inequality_compatibility_numerical_nmat = self.calculateBooleanCompatibilityVector(
                  self.numerical_inequality_nmat,
                  target.numerical_inequality_nmat, is_equality=False)
            if self.bitwise_inequality_nmat != NULL_NMAT:
                inequality_bitwise_arr, inequality_compatibility_bitwise_nmat = self.calculateBooleanCompatibilityVector(
                  self.bitwise_inequality_nmat,
                  target.bitwise_inequality_nmat, is_equality=False, is_bitwise=True)
                inequality_compatibility_arr &= inequality_bitwise_arr
        else:
            is_inequality_compatibility = False
        # Calculate the compatibility vector by combining the equality and inequality constraints
        if is_equality_compatibility and is_inequality_compatibility:
            compatibility_arr = equality_compatibility_arr & inequality_compatibility_arr
        elif is_equality_compatibility:
            compatibility_arr = equality_compatibility_arr
        elif is_inequality_compatibility:
            compatibility_arr = inequality_compatibility_arr
        else:
            # No constraints
            compatibility_arr = np.repeat(True, target.num_row*self.num_row)
        # Create the compatibility collection
        compatibility_collection = CompatibilityCollection(self.num_row, target.num_row)
        target_arr = np.array(range(target.num_row))
        for irow in range(self.num_row):
            #  Select the rows in other that are compatible with the row in self
            base_pos = irow*target.num_row
            idxs = np.array(range(base_pos, base_pos+target.num_row))
            sel_idxs = compatibility_arr[idxs]
            target_rows = target_arr[sel_idxs]
            compatibility_collection.add(irow, target_rows)
        if is_diagnostic:
            result = CompatibilityCollectionResult(compatibility_collection=compatibility_collection,
                    equality_compatibility_nmat=equality_compatibility_nmat,
                    inequality_compatibility_numerical_nmat=inequality_compatibility_numerical_nmat,
                    inequality_compatibility_bitwise_nmat=inequality_compatibility_bitwise_nmat,
                    reference_equality_nmat=self.equality_nmat,
                    target_equality_nmat=target.equality_nmat,
                    reference_numerical_inequality_nmat=self.numerical_inequality_nmat,
                    target_numerical_inequality_nmat=target.numerical_inequality_nmat,
                    reference_bitwise_inequality_nmat=self.bitwise_inequality_nmat,
                    target_bitwise_inequality_nmat=target.bitwise_inequality_nmat)
        else:
            result = CompatibilityCollectionResult(compatibility_collection=compatibility_collection)
        return result
    
    def makeSuccessorPredecessorConstraintMatrix(self)->NamedMatrix:
        """Make a marix of count of successor and predecessor counts. Uses
        subclass.one_step_arr to calculate the counts.

        Returns:
            NamedMatrix: Rows are reactions, columns are constraints by count of reaction type.
              <ReactionType>
        """
        NUM_TRAVERSAL = 2
        def makeTraversalCounts(one_step_arr):
            # one_step_arr: N X N where row is current and column is next
            # Returns: N X NUM_TRAVERSAL where row is current and column is count of next
            multistep_arr = one_step_arr.copy()
            # Process the steps
            count_arrs:list = []  # type: ignore
            count_arrs.append(multistep_arr.sum(axis=1).astype(int))
            for _ in range(2, NUM_TRAVERSAL+1):
                multistep_arr = np.sign(np.matmul(multistep_arr, one_step_arr)).astype(int)
                count_arrs.append(multistep_arr.sum(axis=1))
            count_arr = np.array(count_arrs).T
            return count_arr.astype(int)
        #####
        column_names = [f"s_{n+1}" for n in range(NUM_TRAVERSAL)]
        column_names.extend([f"p_{n+1}" for n in range(NUM_TRAVERSAL)])
        # Calculate successor and predecessor arrays
        successor_count_arr = makeTraversalCounts(self.one_step_nmat.values)
        predecessor_count_arr = makeTraversalCounts(self.one_step_nmat.values.T)
        count_arr = np.concatenate([successor_count_arr, predecessor_count_arr], axis=1).astype(int)
        # Make the NamedMatrix
        named_matrix = NamedMatrix(count_arr, row_names=self.one_step_nmat.row_names,
                           row_description=self.one_step_nmat.row_description,
                           column_description="constraints",
                           column_names=column_names)
        return named_matrix
    
    def makeNStepConstraintMatrix(self, num_step:int=1)->NamedMatrix:
        """Make constraints for a one step navigation of the bipartite graph. For species, the successor constraint
        is the count of reaction types reachable from the species as a reactant via a product of a reaction. For
        reactions, the successor constraint is the count of reaction types reachable from the reactions that produce
        Converseley, the predecessor constraint is the count of reaction types that reachable from previous
        species/reaction.

        Args:
            num_step (int): Number of steps to consider.

        Returns:
            NamedMatrix: Rows are species, columns are constraints
        """
        #####
        def makeStepArr(one_step_arr:np.ndarray)->np.ndarray:
            arrays = []
            for idx in range(num_step):
                matrix_power = np.sign(np.linalg.matrix_power(one_step_arr, idx+1))
                #reactant_to_N_product_arr = np.matmul(matrix_power, self.reactant_nmat.values)
                reactant_to_N_product_arr = np.matmul(matrix_power, self.to_reaction_arr)
                forward_classification_arr = np.matmul(reactant_to_N_product_arr, reaction_to_classification_nmat.values)
                arrays.append(forward_classification_arr)
            return np.concatenate(arrays, axis=1)
        #####
        reaction_to_classification_nmat = ReactionClassification.makeReactionClassificationMatrix(
              self.reactant_nmat.column_names, self.reaction_classification_arr)
        classification_names = reaction_to_classification_nmat.column_names
        reactant_to_product_arr = np.matmul(self.reactant_nmat.values, self.product_nmat.values.T)
        forward_step_arr = makeStepArr(self.one_step_nmat.values)
        backward_step_arr = makeStepArr(self.one_step_nmat.values.T)
        # Column names
        successor_column_names = [f"s{n+1}_{c}" for n in range(num_step) for c in classification_names]
        predecessor_column_names = [f"p{n+1}_{c}" for n in range(num_step) for c in classification_names]
        column_names = successor_column_names + predecessor_column_names
        array = np.concatenate([forward_step_arr, backward_step_arr], axis=1)
        # Construct a matrix where the columns are reaction classifications and the rows are reactions
        named_matrix = NamedMatrix(array, row_names=self.row_names,
                           row_description=self.description,
                           column_description='constraints',
                           column_names=column_names)
        return named_matrix
    
#####################################
NULL_CONSTRAINT = Constraint(NULL_NMAT, NULL_NMAT)


from pySubnetSB.constraint import NULL_NMAT  # type: ignore
from pySubnetSB.species_constraint import SpeciesConstraint  # type: ignore
from pySubnetSB.named_matrix import NamedMatrix # type: ignore
from pySubnetSB.network import Network # type: ignore

import numpy as np
import re
import unittest


IGNORE_TEST = False
IS_PLOT = False
REACTANT_MATRIX = NamedMatrix(np.array([[1, 0], [0, 1], [0, 0]]))
PRODUCT_MATRIX = NamedMatrix(np.array([[1, 1], [1, 0], [0, 0]]))
NUM_ITERATION = 10



#############################
# Tests
#############################
class TestSpeciesConstraint(unittest.TestCase):

    def setUp(self):
        self.constraint = SpeciesConstraint(REACTANT_MATRIX.copy(), PRODUCT_MATRIX.copy()) 

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(self.constraint.reactant_nmat, REACTANT_MATRIX)
        self.assertEqual(self.constraint.product_nmat, PRODUCT_MATRIX)
        self.assertEqual(self.constraint._categorical_nmat, NULL_NMAT)
        self.assertEqual(self.constraint._numerical_enumerated_nmat, NULL_NMAT)

    def testMakeSpeciesConstraintMatrixScale(self):
        if IGNORE_TEST:
            return
        for _ in range(10):
            size = 20
            network = Network.makeRandomNetworkByReactionType(size, size)
            uni_result =  re.findall(": S. ->", str(network))
            num_uni_re = len(list(uni_result))
            uni_result =  re.findall(": S.. ->", str(network))
            num_uni_re += len(list(uni_result))
            species_constraint = SpeciesConstraint(network.reactant_nmat, network.product_nmat)
            named_matrix = species_constraint._makeReactantProductConstraintMatrix()
            self.assertTrue(isinstance(named_matrix, NamedMatrix))
            df = named_matrix.dataframe
            uni_names = ['r_uni-null', 'r_uni-uni', 'r_uni-bi', 'r_uni-multi']
            num_uni_nmat = df[uni_names].sum().sum()
            self.assertEqual(num_uni_re, num_uni_nmat)

    def testMakeAutocatalysisConstraint(self):
        if IGNORE_TEST:
            return
        for _ in range(30):
            size = 20
            network = Network.makeRandomNetworkByReactionType(size, size)
            collection = []
            for stg in [" ", "\n", "$"]:
                totals = [len(re.findall("S" + str(n) + " .*->.*S" + str(n) + stg, str(network)))
                      for n in range(size)]
                collection.append(totals)
            total_arr = np.sign(np.sum(np.array(collection), axis=0))
            species_constraint = SpeciesConstraint(network.reactant_nmat, network.product_nmat)
            named_matrix = species_constraint._makeAutocatalysisConstraint()
            self.assertLessEqual(np.sum(total_arr), np.sum(named_matrix.values))

    def testSpeciesConstraintMatrix(self):
        if IGNORE_TEST:
            return
        for _ in range(30):
            size = 20
            network = Network.makeRandomNetworkByReactionType(size, size)
            species_constraint = SpeciesConstraint(network.reactant_nmat, network.product_nmat)
            named_matrix = species_constraint._makeAutocatalysisConstraint()
            self.assertTrue(isinstance(named_matrix, NamedMatrix))
            df = named_matrix.dataframe
            self.assertGreater(len(df), 0)

    def testCategoricalAndEnumeratedConstraints(self):
        if IGNORE_TEST:
            return
        for _ in range(4):
            self.constraint.setSubset(True)
            self.assertTrue(self.constraint.equality_nmat is NULL_NMAT)
            self.assertTrue(self.constraint.numerical_inequality_nmat is not NULL_NMAT)
            #
            self.constraint.setSubset(False)
            self.assertTrue(self.constraint.equality_nmat is not NULL_NMAT)
            self.assertTrue(self.constraint.numerical_inequality_nmat is NULL_NMAT)

    def testmakeCompatibilityCollection(self):
        if IGNORE_TEST:
            return
        num_permutations = []
        for _ in range(100):
            reference_size = 15
            filler_size = 5*reference_size
            network = Network.makeRandomNetworkByReactionType(reference_size, reference_size)
            big_network = network.fill(num_fill_reaction=filler_size, num_fill_species=filler_size)
            reaction_constraint = SpeciesConstraint(network.reactant_nmat, network.product_nmat,
                                                   is_subnet=True)
            big_reaction_constraint = SpeciesConstraint(big_network.reactant_nmat, big_network.product_nmat,
                                                       is_subnet=True)
            compatibility_collection = reaction_constraint.makeCompatibilityCollection(
                  big_reaction_constraint).compatibility_collection
            name_arr = np.array(big_reaction_constraint.reactant_nmat.row_names)
            for i, arr in enumerate(compatibility_collection.compatibilities):
                reference_name = "S" + str(i)
                target_names = [name_arr[i] for i in arr]
                self.assertTrue(reference_name in target_names)
            num_permutations.append(compatibility_collection.log10_num_assignment)
        #print(np.mean(num_permutations))

    def testMakeBitwiseReactantProductConstraintMatrix(self):
        if IGNORE_TEST:
            return
        for _ in range(NUM_ITERATION):
            size = 10
            network = Network.makeRandomNetworkByReactionType(size, size)
            species_constraint = SpeciesConstraint(network.reactant_nmat, network.product_nmat)
            named_matrix = species_constraint._makeBitwiseReactantProductConstraintMatrix()
            self.assertTrue(isinstance(named_matrix, NamedMatrix))
            df = named_matrix.dataframe
            self.assertGreater(len(df), 0)

    def testMakeReactantProductCountConstraintMatrix(self):
        if IGNORE_TEST:
            return
        for _ in range(NUM_ITERATION):
            size = 10
            network = Network.makeRandomNetworkByReactionType(size, size)
            species_constraint = SpeciesConstraint(network.reactant_nmat, network.product_nmat)
            named_matrix = species_constraint._makeReactantProductCountConstraintMatrix()
            count_arr = np.sum(named_matrix.values, axis=1)
            for idx in range(len(count_arr)):
                scanned_count = str(network).count("S" + str(idx))
                self.assertEqual(count_arr[idx], scanned_count)
            df = named_matrix.dataframe
            self.assertGreater(len(df), 0)

    def testMakeSuccessorPredecessorConstraintMatrix(self):
        if IGNORE_TEST:
            return
        antimony = """
        J0:  -> S2
        J1: S2 -> S0
        J2: S0 + S1 -> S1
        J3: S0 -> S0
        """
        network = Network.makeFromAntimonyStr(antimony)
        species_constraint = SpeciesConstraint(network.reactant_nmat, network.product_nmat)
        named_matrix = species_constraint.makeNStepConstraintMatrix()
        df = named_matrix.dataframe
        ser = df.loc["S0", :]
        result_dct = {"s1_uni-uni": 1, "s1_bi-uni": 2, "p1_uni-uni": 2, "p1_bi-uni": 1}
        for idx in result_dct.keys():
            self.assertEqual(ser[idx], result_dct[idx])
        other_index = set(ser.index) - set(result_dct.keys())
        trues = [ser[idx] == 0 for idx in other_index]
        self.assertTrue(all(trues))
    
    def testMakeSuccessorPredecessorConstraintMatrixScale(self):
        if IGNORE_TEST:
            return
        antimony = """
        J0:  -> S2
        J1: S2 -> S0
        J2: S0 + S1 -> S1
        J3: S0 -> S0
        """
        factor = 5
        network = Network.makeFromAntimonyStr(antimony)
        species_constraint = SpeciesConstraint(network.reactant_nmat, network.product_nmat)
        named_matrix_one = species_constraint.makeNStepConstraintMatrix(num_step=1)
        named_matrix_factor = species_constraint.makeNStepConstraintMatrix(num_step=factor)
        self.assertEqual(len(named_matrix_factor.column_names), factor*len(named_matrix_one.column_names))
        self.assertEqual(len(named_matrix_factor.row_names), len(named_matrix_one.row_names))


if __name__ == '__main__':
    unittest.main()


'''Mocks the multiprocessing Queue to handle single process execution.'''


class MockQueue(object):
    # Mocks Multiprocessing Queue
    def __init__(self):
        self.queue = []

    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

    def put(self, item):
        self.queue.insert(0, item)

    def get(self):
        return self.queue.pop()
    
    def empty(self):
        return len(self.queue) == 0
    
    def qsize(self):
        return len(self.queue)
    
    def close(self):
        pass

    def task_done(self):
        pass

    def join(self):
        pass


'''Manage the distribution of work for SubnetFinder workers.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.network import Network # type: ignore
from pySubnetSB.model_serializer import ModelSerializer # type: ignore

import collections  # type: ignore
import itertools
import pandas as pd  # type: ignore
import numpy as np # type: ignore
import os
from typing import List, Optional, Tuple


Workunit = collections.namedtuple("Workunit", ['reference_idx', 'target_idx', 'worker_idx',
      'reference_name', 'target_name'])


class SubnetFinderWorkunitManager(object):

    def __init__(self, csv_path:str, num_worker:Optional[int]=None,
          reference_networks:Optional[List[Network]]=None,
          target_networks:Optional[List[Network]]=None):
        """
        Args:
            path (str): path to the CSV file used to store the workunits.
            num_worker (Optional[int], optional): Number of workers. Defaults to None.
            reference_networks (Optional[List[Network]], optional): _description_. Defaults to None.
            target_networks (Optional[List[Network]], optional): _description_. Defaults to None.
        """
        self.num_worker = num_worker
        self.reference_networks = reference_networks
        self.target_networks = target_networks
        self.num_reference_network = None if reference_networks is None else len(reference_networks)
        self.num_target_network = None if target_networks is None else len(target_networks)
        self.csv_path = csv_path

    @classmethod
    def makeFromSerializationFiles(cls, csv_path:str, num_worker:int,
          reference_serialization_path:str,
          target_serialization_path:str)->'SubnetFinderWorkunitManager':
        """
        Create a workunit manager from serialization files.

        Args:
            reference_serialization_path (str): Path to the reference networks.
            target_serialization_path (str): Path to the target networks.
        """
        serializer = ModelSerializer(None, reference_serialization_path)
        reference_networks = serializer.deserialize().networks
        serializer = ModelSerializer(None, target_serialization_path)
        target_networks = serializer.deserialize().networks
        return cls(csv_path, num_worker=num_worker,
              reference_networks=reference_networks, target_networks=target_networks)

    def makeWorkunitFile(self, reference_serialization_path:Optional[str]=None,
          target_serialization_path:Optional[str]=None)->None:
        """
        Creates a CSV file of workunits.
        Randomly assigns workers to work units.
        Make workunit pairs for each worker.

        Args:
            reference_serialization_path (Optional[str], optional): Path to the reference networks. Defaults to None.
            target_serialization_path (Optional[str], optional): Path to the target networks. Defaults
        """
        #####
        def getNetworks(serialization_path:Optional[str],
              default_networks:Optional[List[Network]])->Optional[List[Network]]:
            if serialization_path is None:
                return default_networks
            # Deserialize and return the networks
            serializer = ModelSerializer(None, serialization_path)
            return serializer.deserialize().networks
        #####
        reference_networks = getNetworks(reference_serialization_path, self.reference_networks)
        target_networks = getNetworks(target_serialization_path, self.target_networks)
        if reference_networks is None or target_networks is None:
            raise ValueError("Reference or target networks are not set.")
        num_reference_network = len(reference_networks)
        num_target_network = len(target_networks)
        num_workunit_pairs = num_reference_network * num_target_network
        network_idx_pairs = [(i, j) for i, j in itertools.product(
              range(num_reference_network), range(num_target_network))]
        worker_idxs = np.array(range(num_workunit_pairs)) % self.num_worker
        np.random.shuffle(worker_idxs)   # Randomize the assignments
        reference_names = [reference_networks[w[0]].network_name for w in network_idx_pairs]
        target_names = [target_networks[w[1]].network_name for w in network_idx_pairs]
        df = pd.DataFrame({cn.FINDER_REFERENCE_IDX: [network_idx_pairs[i][0] for i in range(len(worker_idxs))],
              cn.FINDER_TARGET_IDX: [network_idx_pairs[i][1] for i in range(len(worker_idxs))],
              cn.FINDER_WORKER_IDX: worker_idxs,
              cn.FINDER_REFERENCE_NAME: reference_names,
              cn.FINDER_TARGET_NAME: target_names})
        df.to_csv(self.csv_path, index=False)

    def getWorkunitDataframe(self, worker_idx:Optional[int]=None)->pd.DataFrame:
        if not os.path.exists(self.csv_path):
            raise ValueError(f"File does not exist: {self.csv_path}.")
        df = pd.read_csv(self.csv_path)
        if worker_idx is not None:
            worker_df = df[df[cn.FINDER_WORKER_IDX] == worker_idx]
        else:
            worker_df = df
        return worker_df
    
    def getWorkunits(self, worker_idx:int,
          processed_reference_networks:Optional[List[str]]=None,
          processed_target_networks:Optional[List[str]]=None,
          )->List[Workunit]:
        """
        Get the workunits for a worker, eliminating any workunits that have been processed.

        Args:
            worker_idx (int): index of the worker
            processed_reference_networks (Optional[List[str]], optional): List of names of processed
                reference networks paired with processed_target_networks to define the workunit.
            processed_target_networks (Optional[List[str]], optional): List of names of processed target networks
                paired with processed_reference_networks to define the workunit

        Returns:
            List[Workunit]
        """
        worker_df = self.getWorkunitDataframe(worker_idx=worker_idx)
        if processed_reference_networks is None:
            processed_reference_networks = []
        if processed_target_networks is None:
            processed_target_networks = []
        if len(processed_reference_networks) != len(processed_target_networks):
            raise ValueError("Length of processed reference and target networks must be equal.")
        #
        if len(processed_reference_networks) > 0:
            # Eliminate previously processed workunits
            sel = worker_df[cn.FINDER_REFERENCE_NAME].isin(processed_reference_networks)
            sel = np.logical_and (sel, worker_df[cn.FINDER_TARGET_NAME].isin(processed_target_networks))
            worker_df = worker_df[~sel]   # Remove the processed workunits
        workunits = [Workunit(reference_idx=row[cn.FINDER_REFERENCE_IDX], target_idx=row[cn.FINDER_TARGET_IDX], 
              worker_idx=worker_idx, reference_name=row[cn.FINDER_REFERENCE_NAME],
              target_name=row[cn.FINDER_TARGET_NAME]) for _, row in worker_df.iterrows()]
        return workunits
    
    def remove(self):
        """Remove the workunit file."""
        if os.path.exists(self.csv_path):
            os.remove(self.csv_path)


from pySubnetSB.named_matrix import NamedMatrix  # type: ignore

import copy
import numpy as np
import unittest
import time


IGNORE_TEST = False
IS_PLOT = False
MAT = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
MAT2 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0, 0]])


#############################
# Tests
#############################
class TestNamedMatrix(unittest.TestCase):

    def setUp(self):
        self.array = copy.copy(MAT)
        self.named_matrix = NamedMatrix(MAT.copy(), row_names=[(1, 0), (0, 1), (0, 0)],
                                        column_names=['d', 'e', 'f'])

    def testConstructor(self):
        if IGNORE_TEST:
            return
        row_description = "rows"
        column_description = "columns"
        named_matrix = NamedMatrix(MAT2.copy(), row_names=[(1, 0), (0, 1), (0, 0), (0, 3)],
                                        column_names=['d', 'e', 'f'],
                                    row_description=row_description, column_description=column_description)
        self.assertTrue(np.all(named_matrix.values == MAT2))
        self.assertTrue("e" in str(named_matrix))
        self.assertTrue(row_description in str(named_matrix))
        self.assertTrue(column_description in str(named_matrix))

    def testPerformance0(self):
        if IGNORE_TEST:
            return
        arr = np.random.randint(-10, 10, (100, 100))
        start_time = time.time()
        for _ in range(10000):
            named_matrix = NamedMatrix(arr)
        elapsed_time = time.time() - start_time
        self.assertLess(elapsed_time, 1e-1)

    def testEq(self):
        if IGNORE_TEST:
            return
        self.assertTrue(self.named_matrix == self.named_matrix)
        #
        named_matrix = NamedMatrix(MAT2.copy(), row_names=[(1, 0), (0, 1), (0, 0), (0, 3)],
                                        column_names=['d', 'e', 'f'])
        self.assertFalse(self.named_matrix == named_matrix)

    def testLe(self):
        if IGNORE_TEST:
            return
        self.assertTrue(self.named_matrix <= self.named_matrix)
        #
        named_matrix = NamedMatrix(MAT2.copy(), row_names=[(1, 0), (0, 1), (0, 0), (0, 3)],
                                        column_names=['d', 'e', 'f'])
        self.assertFalse(self.named_matrix <= named_matrix)
        #
        mat = MAT.copy()
        mat[0, 0] = 10
        named_matrix = NamedMatrix(mat, row_names=[(1, 0), (0, 1), (0, 0)],
                                        column_names=['d', 'e', 'f'])
        self.assertFalse(named_matrix <= self.named_matrix)

    def testTemplate(self):
        if IGNORE_TEST:
            return
        named_matrix = self.named_matrix.template()
        self.assertTrue(named_matrix == self.named_matrix)
        #
        mat = MAT.copy()
        mat[0, 0] = 10
        named_matrix = self.named_matrix.template(matrix=mat)
        self.assertFalse(named_matrix == self.named_matrix)
        self.assertTrue(named_matrix.isCompatible(self.named_matrix))

    def testGetSubNamedMatrix(self):
        if IGNORE_TEST:
            return
        result = self.named_matrix.getSubNamedMatrix(row_names=[(1, 0), (0, 1)], column_names=['d', 'e'])
        named_matrix = result.named_matrix
        self.assertTrue(np.all(named_matrix.values == np.array([[1, 0], [0, 1]])))
        self.assertTrue(np.all(named_matrix.row_names == np.array([(1, 0), (0, 1)])))
        self.assertTrue(np.all(named_matrix.column_names == np.array(['d', 'e'])))

    def testGetSubMatrix(self):
        if IGNORE_TEST:
            return
        matrix = self.named_matrix.getSubMatrix(row_idxs=range(2), column_idxs=range(2))
        self.assertTrue(np.all(matrix.values == np.array([[1, 0], [0, 1]])))

    def testPerformance(self):
        if IGNORE_TEST:
            return
        num_row = 100
        num_column = 100
        mat = np.random.randint(-10, 10, (num_row, num_column))
        named_matrix = NamedMatrix(mat)
        def timeit(num_iteration=1000, is_named_matrix=True):
            t0 = time.time()
            for _ in range(num_iteration):
                if is_named_matrix:
                    _ = named_matrix.getSubNamedMatrix(row_names=range(10), column_names=range(10))
                else:
                    _ = named_matrix.getSubMatrix(row_idxs=range(10), column_idxs=range(10))
            t1 = time.time()
            return t1 - t0
        #
        time_named_matrix = timeit(is_named_matrix=True)
        time_matrix = timeit(is_named_matrix=False)
        self.assertTrue(time_named_matrix/time_matrix > 100)
        #print(f"TimeNamed: {time_named_matrix}", f"TimeMatrix: {time_matrix}")

    def testCopyEquals(self):
        if IGNORE_TEST:
            return
        named_matrix = self.named_matrix.copy()
        self.assertTrue(named_matrix == self.named_matrix)

    def testTranspose(self):
        if IGNORE_TEST:
            return
        named_matrix = self.named_matrix.transpose()
        reverted = named_matrix.transpose()
        self.assertTrue(reverted == self.named_matrix)

    def testSerializeDeserialize(self):
        if IGNORE_TEST:
            return
        string = self.named_matrix.serialize()
        named_matrix = NamedMatrix.deserialize(string)
        self.assertTrue(named_matrix == self.named_matrix)

    def testHstack(self):
        if IGNORE_TEST:
            return
        named_matrix = NamedMatrix(np.array([[1, 0], [0, 1], [0, 0]]), column_names=['a', 'b'])
        named_matrix2 = NamedMatrix(np.array([[1], [0], [0]]), column_names=['c'])
        # Same shapes
        result = NamedMatrix.hstack([named_matrix, named_matrix2])
        expected_arr = np.array([[1, 0, 1], [0, 1, 0], [0, 0, 0]])
        self.assertTrue(np.all(result.values == expected_arr))
        self.assertTrue(np.all(result.column_names == np.array(['a', 'b', 'c'])))
        self.assertEqual(result.num_column, named_matrix.num_column + named_matrix2.num_column)
        # Different shapes
        with self.assertRaises(ValueError):
            NamedMatrix.hstack([named_matrix, named_matrix2.transpose()])
        # Different row names
        with self.assertRaises(ValueError):
            named_matrix2 = NamedMatrix(np.array([[1], [0], [0]]), column_names=['c'],
                                        row_names=['x', 'y', 'z'])
            _ = NamedMatrix.hstack([named_matrix, named_matrix2])

    def testVstack(self):
        if IGNORE_TEST:
            return
        array1 =np.array([[1, 0], [0, 1], [0, 0]])
        array2 =np.array([[1, 2], [0, 2]])
        named_matrix1 = NamedMatrix(array1, row_names=['a', 'b', 'c'])
        named_matrix2 = NamedMatrix(array2, row_names=['d', 'e'])
        # Same shapes
        result = NamedMatrix.vstack([named_matrix1, named_matrix2])
        self.assertTrue(np.all(result.values == np.vstack([array1, array2])))
        self.assertTrue(np.all(result.row_names == np.array(['a', 'b', 'c', 'd', 'e'])))
        self.assertEqual(result.num_row, named_matrix1.num_row + named_matrix2.num_row)
        # Different shapes
        with self.assertRaises(ValueError):
            NamedMatrix.vstack([named_matrix1, named_matrix2.transpose()])
        # Different row names
        with self.assertRaises(ValueError):
            named_matrix2 = NamedMatrix(np.array([[1], [0], [0]]),
                                        column_names=['x', 'y'])
            _ = NamedMatrix.hstack([named_matrix1, named_matrix2])

    def testVmerge(self):
        if IGNORE_TEST:
            return
        named_matrix1 = NamedMatrix(np.array([[1, 0], [0, 1], [1, 1]]),
              row_names=['x', 'y', 'z'], column_names=['a', 'b'])
        named_matrix2 = NamedMatrix(np.array([[1, 0], [0.5, 0.5], [1, 1]]),
              row_names=['x', 'w', 'z'], column_names=['c', 'd'])
        named_matrix = named_matrix1.vmerge(named_matrix2)
        self.assertEqual(named_matrix.num_row, 4)
        self.assertEqual(named_matrix.num_column, 4)
    
    def testVmerge1(self):
        if IGNORE_TEST:
            return
        num_overlap = 5
        size = 10
        merge_size = 2*size - num_overlap
        named_matrix1 = NamedMatrix.makeRandom(size, size)
        named_matrix2 = NamedMatrix.makeRandom(size, size)
        column_names = ["C" + n for n in named_matrix1.column_names]
        row_names = ["R" + n if int(n) < num_overlap else n for n in named_matrix1.row_names]
        named_matrix2 = NamedMatrix(named_matrix2.values, row_names=row_names, column_names=column_names)
        named_matrix = named_matrix1.vmerge(named_matrix2)
        self.assertEqual(named_matrix.num_row, 2*size - num_overlap)
        self.assertEqual(named_matrix.num_column, 2*size)
        # Check that the values are correct
        # There are 3 "bands" of rows: 
        #  (a) 0's in the matrix2-only columns; 
        #  (b) all non-zero; 
        #  (c) 0's in the matrix1-only columns
        type_a_rows = list(range(0, size-num_overlap))
        type_b_rows = list(range(size-num_overlap, size))
        matrix1_columns = list(range(size))
        matrix2_columns = list(range(size, 2*size))
        for irow in range(merge_size):
            if irow in type_a_rows:
                self.assertTrue(np.all(named_matrix.values[irow, np.array(matrix1_columns)] != 0))
                self.assertTrue(np.all(named_matrix.values[irow, np.array(matrix2_columns)] == 0))
            elif irow in type_b_rows:
                self.assertTrue(np.all(named_matrix.values[irow, np.array(matrix1_columns)] != 0))
                self.assertTrue(np.all(named_matrix.values[irow, np.array(matrix2_columns)] != 0))
            else:
                self.assertTrue(np.all(named_matrix.values[irow, np.array(matrix1_columns)] == 0))
                self.assertTrue(np.all(named_matrix.values[irow, np.array(matrix2_columns)] != 0))

    def testRandomizeAndSort(self):
        if IGNORE_TEST:
            return
        named_matrix = NamedMatrix.makeRandom(10, 10)
        sorted_named_matrix = named_matrix.sort()
        randomized_named_matrix = named_matrix.randomize().named_matrix
        sorted_randomized_named_matrix = randomized_named_matrix.sort()
        self.assertEqual(sorted_named_matrix, sorted_randomized_named_matrix)

        

if __name__ == '__main__':
    unittest.main(failfast=True)


'''Benchmarks for isStructurallyIdentical and isStructurallyIdenticalSubnet.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.network import Network # type: ignore
from pySubnetSB.assignment_pair import AssignmentPair  # type: ignore

import json
import numpy as np
import os
import shutil
import time
from typing import List


REFERENCE = "reference"
TARGET = "target"
S_RUNTIMES = "s_runtimes"
S_NUM_SUCCESS = "s_num_success"
S_NUM_TRUNCATED = "s_num_truncated"
S_IS_IDENTICALS = "s_is_identicals"
S_IS_TRUNCATEDS = "s_is_truncateds"
S_REFERENCE_SIZE = "s_reference_size"
S_EXPANSION_FACTOR = "s_expansion_factor"
S_NUM_EXPERIMENT = "s_num_experiment"
S_IS_IDENTICAL = "s_is_identical"
S_EXPERIMENTS = "s_experiments"
S_BENCHMARK_RUNNER = "s_benchmark_runner"


##############################
class Experiment(object):

    def __init__(self, reference:Network, target:Network, assignment_pair:AssignmentPair):
        self.reference = reference
        self.target = target
        self.assignment_pair = assignment_pair

    def __eq__(self, other)->bool:
        if not isinstance(other, Experiment):
            return False
        if not self.reference == other.reference:
            return False
        if not self.target == other.target:
            return False
        if not self.assignment_pair == other.assignment_pair:
            return False
        return True

    def __repr__(self):
        #####
        def networkRepr(network:Network):
            return f"(num_species: {network.num_species}, num_reaction: {network.num_reaction})"
        #####
        return f"reference: {networkRepr(self.reference)}, target: {networkRepr(self.target)}"
    
    def serialize(self):
        """Create a JSON string for the object.

        Returns:
            str: _description_
        """
        dict = {cn.S_ID: self.__class__.__name__,
                cn.S_REFERENCE: self.reference.serialize(), cn.S_TARGET: self.target.serialize(),
                cn.S_ASSIGNMENT_PAIR: self.assignment_pair.serialize()}
        return json.dumps(dict)
    
    @classmethod
    def deserialize(cls, json_str:str)->'Experiment':
        """Create an object from a JSON string.

        Args:
            json_str (str): _description_

        Returns:
            Experiment: _description_
        """
        dict = json.loads(json_str)
        reference = Network.deserialize(dict[cn.S_REFERENCE])
        target = Network.deserialize(dict[cn.S_TARGET])
        assignment_pair = AssignmentPair.deserialize(dict[cn.S_ASSIGNMENT_PAIR])
        return Experiment(reference=reference, target=target, assignment_pair=assignment_pair)
    

##############################
class ExperimentResult(object):

    def __init__(self, benchmark_runner:'BenchmarkRunner',
                 num_experiment:int=0, runtimes:List[float]=[], num_success:int=0,
          num_truncated:int=0, is_identicals:List[bool]=[], is_truncateds:List[bool]=[]):
        self.benchmark_runner = benchmark_runner
        self.num_experiment = num_experiment
        self.runtimes = runtimes
        self.num_success = num_success
        self.num_truncated = num_truncated
        self.is_identicals = is_identicals
        self.is_truncateds = is_truncateds

    def serialize(self)->str:
        """Create a JSON string for the object.

        Returns:
            str: _description_
        """
        dict = {cn.S_ID: self.__class__.__name__,
                S_BENCHMARK_RUNNER: self.benchmark_runner.serialize(),
                S_NUM_EXPERIMENT: self.num_experiment,
                S_RUNTIMES: self.runtimes, S_NUM_SUCCESS: self.num_success,
                S_NUM_TRUNCATED: self.num_truncated, S_IS_IDENTICALS: self.is_identicals,
                S_IS_TRUNCATEDS: self.is_truncateds}
        return json.dumps(dict)
    
    @classmethod
    def deserialize(cls, json_str:str)->'ExperimentResult':
        """Create an object from a JSON string.

        Args:
            json_str (str): _description_

        Returns:
            Experiment: _description_
        """
        dict = json.loads(json_str)
        benchmark_runner = BenchmarkRunner.deserialize(dict[S_BENCHMARK_RUNNER])
        num_experiment = dict[S_NUM_EXPERIMENT]
        runtimes = dict[S_RUNTIMES]
        num_success = dict[S_NUM_SUCCESS]
        num_truncated = dict[S_NUM_TRUNCATED]
        is_identicals = dict[S_IS_IDENTICALS]
        is_truncateds = dict[S_IS_TRUNCATEDS]
        return ExperimentResult(benchmark_runner, num_experiment=num_experiment,
                runtimes=runtimes, num_success=num_success, num_truncated=num_truncated,
                is_identicals=is_identicals, is_truncateds=is_truncateds)

    def __eq__(self, other)->bool:
        if not isinstance(other, ExperimentResult):
            return False
        if self.num_experiment != other.num_experiment:
            return False
        if not np.all(self.runtimes == other.runtimes):
            return False
        if self.num_success != other.num_success:
            return False
        if self.num_truncated != other.num_truncated:
            return False
        if not np.all(self.is_identicals == other.is_identicals):
            return False
        if not np.all(self.is_truncateds == other.is_truncateds):
            return False
        return True

    def __repr__(self):
        frac_truncated = self.num_truncated/self.num_experiment
        frac_success = self.num_success/self.num_experiment
        return f"num_experiment: {self.num_experiment}, frac_success: {frac_success}, " + \
                f"frac_truncated: {frac_truncated}, avg_runtime: {np.mean(self.runtimes)}"
    
    def getIdenticals(self)->List[Experiment]:
        """
        Gets the experiments that found an identical subnetwork.

        Returns:
            List[Experiment]
        """
        return [e for i, e in enumerate(self.benchmark_runner.experiments) if self.is_identicals[i]] 

    def getTruncateds(self)->List[Experiment]:
        """
        Gets the experiments whose search was truncated.

        Returns:
            List[Experiment]
        """
        return [e for i, e in enumerate(self.benchmark_runner.experiments) if self.is_truncateds[i]]


##############################
class BenchmarkRunner(object):

    def __init__(self, 
                num_experiment:int=10, 
                reference_size:int=3,
                expansion_factor:int=1,
                is_identical:bool=True,
                identity=cn.ID_WEAK):
        """
        Args:
            num_experiment (int): number of experiments in the benchmark
            reference_size (int)
            expansion_factor (int): Integer factor for size of target relative to reference
            is_identical (bool): If True, the target has a subnet equivalent to the reference.
            identity (str)
        """
        self.reference_size = reference_size
        self.expansion_factor = expansion_factor
        self.identity = identity
        self.num_experiment = num_experiment
        self.is_identical = is_identical
        self.experiments = [self.makeExperiment(is_identical=self.is_identical)]*num_experiment

    def __eq__(self, other)->bool:
        if not isinstance(other, BenchmarkRunner):
            return False
        if self.reference_size != other.reference_size:
            return False
        if self.expansion_factor != other.expansion_factor:
            return False
        if self.identity != other.identity:
            return False
        if self.num_experiment != other.num_experiment:
            return False
        if len(self.experiments) != len(other.experiments):
            return False
        for idx, experiment in enumerate(self.experiments):
            if not experiment == other.experiments[idx]:
                return False
        return True

    def makeExperiment(self, is_identical=True)->Experiment:
        """Construct the elements of an experiment.

        Args:
            is_identical (bool): If True, the target has a subnet equivalent to the reference.

        Returns:
            Experiment
        """
        #####
        def changeValue(value):
            if value == 0:
                return 1
            else:
                return 0
        #####
        reference = Network.makeRandomNetworkByReactionType(self.reference_size, is_prune_species=False)
        target_size = self.reference_size*self.expansion_factor
        filler_size = target_size - self.reference_size
        if filler_size == 0:
            # Do not expand the target
            target, assignment_pair = reference.permute()
        else:
            # Expand the target
            #   Create the left part of the expanded target
            left_filler_arr = np.zeros((filler_size, self.reference_size))
            xreactant_arr = np.vstack([reference.reactant_nmat.values, left_filler_arr])
            xproduct_arr = np.vstack([reference.product_nmat.values, left_filler_arr])
            #  Create the expanded arrays
            right_filler = Network.makeRandomNetworkByReactionType(num_reaction=filler_size,
                  num_species=target_size, is_prune_species=False)
            xreactant_arr = np.hstack([xreactant_arr, right_filler.reactant_nmat.values])
            xproduct_arr = np.hstack([xproduct_arr, right_filler.product_nmat.values])
            xtarget = Network(xreactant_arr, xproduct_arr)
            #  Merge the left and right parts
            target, assignment_pair = xtarget.permute()
            assignment_pair = assignment_pair.resize(self.reference_size)
        if not is_identical:
            i_species, i_reaction = np.random.randint(0, reference.num_species),  \
                  np.random.randint(0, reference.num_reaction)
            max_val = np.max(reference.reactant_nmat.values)
            # Insert an impossible value
            reference.reactant_nmat.values[i_species, i_reaction] =  max_val + 1
            reference.standard_nmat.values = reference.product_nmat.values - reference.reactant_nmat.values
        return Experiment(reference=reference, target=target, assignment_pair=assignment_pair)
    
    def run(self)->ExperimentResult:
        """Run the benchmark.

        Returns:
            ExperimentResult
        """
        runtimes = []
        num_success = 0
        num_truncated = 0
        is_identicals = []
        is_truncateds = []
        for experiment in self.experiments:
            start_time = time.process_time()
            result = experiment.reference.isStructurallyIdentical(experiment.target,
                  identity=self.identity, is_subnet=True)
            runtimes.append(time.process_time() - start_time)
            is_identical = any([ap == experiment.assignment_pair for ap in result.assignment_pairs])
            is_identicals.append(is_identical)
            is_truncateds.append(result.is_truncated)
            if is_identical:
                num_success += 1
            if result.is_truncated:
                num_truncated += 1
        return ExperimentResult(self, num_experiment=self.num_experiment, runtimes=runtimes,
              num_success=num_success, num_truncated=num_truncated,
              is_identicals=is_identicals, is_truncateds=is_truncateds)
    
    def exportExperimentsAsCSV(self, directory:str)->None:
        """Creates two subdirectores, reference and target, with the experiments in CSV format.

        Args:
            directory (str): _description_
        """
        #####
        def mkSubdir(subdir:str):
            subdir = os.path.join(directory, subdir)
            if os.path.exists(subdir):
                shutil.rmtree(subdir)
            os.makedirs(subdir)
            return subdir
        #####
        def writeSubdir(is_reference:bool=True):
            if is_reference:
                subdir = mkSubdir(REFERENCE)
            else:
                subdir = mkSubdir(TARGET)
            #
            for idx, experiment in enumerate(self.experiments):
                if is_reference:
                    network = experiment.reference
                else:
                    network = experiment.target
                network_str = network.makeCSVNetwork(identity=self.identity)
                path = os.path.join(subdir, f"{idx}.txt")
                with open(path, "w") as file:
                    file.write(network_str)
        #####
        # Create the CSV files
        writeSubdir(is_reference=True)
        writeSubdir(is_reference=False)

    def serialize(self)->str:
        """Create a JSON string for the object.

        Returns:
            str: _description_
        """
        experiments = [experiment.serialize() for experiment in self.experiments]
        dict = {cn.S_ID: self.__class__.__name__,
                S_REFERENCE_SIZE: self.reference_size, S_EXPANSION_FACTOR: self.expansion_factor,
                S_NUM_EXPERIMENT: self.num_experiment, S_IS_IDENTICAL: self.is_identical,
                S_EXPERIMENTS: experiments, cn.S_IDENTITY: self.identity}
        return json.dumps(dict)
    
    @classmethod
    def deserialize(cls, serialization_str:str)->'BenchmarkRunner':
        """Create an object from a JSON string.

        Args:
            serialization_str (str): _description_

        Returns:
            BenchmarkRunner: _description_
        """
        dict = json.loads(serialization_str)
        reference_size = dict[S_REFERENCE_SIZE]
        expansion_factor = dict[S_EXPANSION_FACTOR]
        num_experiment = dict[S_NUM_EXPERIMENT]
        is_identical = dict[S_IS_IDENTICAL]
        identity = dict[cn.S_IDENTITY]
        experiments = [Experiment.deserialize(exp_str) for exp_str in dict[S_EXPERIMENTS]]
        benchmark_runner = BenchmarkRunner(num_experiment=num_experiment, reference_size=reference_size,
              expansion_factor=expansion_factor, is_identical=is_identical, identity=identity)
        benchmark_runner.experiments = experiments
        return benchmark_runner


'''Finds subnets for pairs of reference, target networks and constructs a DataFrame of results.'''

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.model_serializer import ModelSerializer # type: ignore
from pySubnetSB.network import Network  # type: ignore
import pySubnetSB.constants as cn
from pySubnetSB.assignment_pair import AssignmentPair # type: ignore

import collections
import json
import os
import numpy as np
import pandas as pd  # type: ignore
from typing import List, Tuple, Optional


NetworkPair = collections.namedtuple("NetworkPair", "reference_network target_network")


############################### CLASSES ###############################
class SubnetFinder(object):

    def __init__(self, network_pairs:List[NetworkPair], identity:str=cn.ID_WEAK,
          num_process:int=-1)->None:
        """
        Args:
            network_pairs (List[NetworkPair]): List of NetworkPair
            identity (str): Identity type
            num_process (int): Number of processes to use. If -1, use all available processors.
        """
        self.network_pairs = network_pairs
        self.identity = identity
        self.num_process = num_process

    @classmethod
    def makeFromCombinations(cls, reference_networks:List[Network], target_networks:List[Network],
                 identity:str=cn.ID_WEAK, num_process:int=-1)->"SubnetFinder":
        """
        Makes a SubnetFinder from the combinations of reference and target networks.
        Args:
            reference_networks (List[Network]): Reference networks
            target_networks (List[Network]): Target networks
            identity (str): Identity type
            num_process (int): Number of processes to use. If -1, use all available processors.

        Returns:
            SubnetFinder
        """
        network_pairs = [NetworkPair(reference_network, target_network)
                         for reference_network in reference_networks
                         for target_network in target_networks]
        return cls(network_pairs, identity=identity, num_process=num_process)

    def find(self, is_report:bool=True, max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT)->pd.DataFrame:
        """
        Finds subnets of SBML/Antmony models in a target directory for SBML/Antimony models in a reference directory.

        Args:
            is_report (bool): If True, report progress
            max_num_assignment (int): Maximum number of assignment pairs

        Returns
            pd.DataFrame: Table of matching networks (cn.FINDER_DATAFRAME_COLUMNS)
                reference_name (str): Reference model name
                target_name (str): Target model name
                reference_network (str): string representation of the reference network
                induced_network (str): string representation of the induced network in the target
                name_dct (dict): Dictionary of mapping of target names to reference names for species and reactions
                                 as a JSON string.
                num_assignment_pair (int): Number of assignment pairs returned
                                            for matching target to subnet
                is_trucated (bool): True if the search is truncated
        """
        dct:dict = {k: [] for k in cn.FINDER_DATAFRAME_COLUMNS}
        for network_pair in self.network_pairs:
            reference_network = network_pair.reference_network
            target_network = network_pair.target_network
            if is_report:
                print(f"Processing reference model: {reference_network.network_name}")
            result = reference_network.isStructurallyIdentical(
                  target_network,
                  identity=self.identity,
                  num_process=self.num_process,
                  is_report=is_report,
                  is_subnet=True,
                  max_num_assignment=max_num_assignment)
            dct[cn.FINDER_REFERENCE_NAME].append(reference_network.network_name)
            dct[cn.FINDER_TARGET_NAME].append(target_network.network_name)
            dct[cn.FINDER_IS_TRUNCATED].append(result.is_truncated)
            if not result:
                dct[cn.FINDER_REFERENCE_NETWORK].append(cn.NULL_STR)
                dct[cn.FINDER_INDUCED_NETWORK].append(cn.NULL_STR)
                dct[cn.FINDER_NUM_ASSIGNMENT_PAIR].append(cn.NULL_STR)
                dct[cn.FINDER_NUM_MAPPING_PAIR].append(cn.NULL_STR)
                dct[cn.FINDER_NAME_DCT].append(cn.NULL_STR)
            else:
                # Construct the induced subnet
                species_assignment_arr = result.assignment_pairs[0].species_assignment
                reaction_assignment_arr = result.assignment_pairs[0].reaction_assignment
                if is_report:
                    print(f"Found matching model: {reference_network.network_name} and {target_network.network_name}")
                dct[cn.FINDER_REFERENCE_NETWORK].append(str(reference_network))
                dct[cn.FINDER_INDUCED_NETWORK].append(str(result.induced_network))
                dct[cn.FINDER_NUM_ASSIGNMENT_PAIR].append(len(result.assignment_pairs))
                dct[cn.FINDER_NUM_MAPPING_PAIR].append(len(result.assignment_pairs))
                # Create a more complete assignment pair
                assignment_pair = AssignmentPair(species_assignment=species_assignment_arr,
                        reaction_assignment=reaction_assignment_arr,
                        reference_reaction_names=reference_network.reaction_names,
                        reference_species_names=reference_network.species_names,
                        target_reaction_names=target_network.reaction_names,
                        target_species_names=target_network.species_names)
                dct_str = json.dumps(assignment_pair.makeNameDct())
                dct[cn.FINDER_NAME_DCT].append(dct_str)
        df = pd.DataFrame(dct)
        return df
    
    @classmethod
    def findFromDirectories(cls, reference_directory, target_directory, identity:str=cn.ID_WEAK,
          is_report:bool=True)->pd.DataFrame:
        """
        Finds subnets of SBML/Antmony models in a target directory for SBML/Antimony models in a reference directory.

        Args:
            reference_directory (str): Directory that contains the reference model files
            target_directory (str): Directory that contains the target model files or a serialization file (".txt")
            identity (str): Identity type
            is_report (bool): If True, report progress

        Returns:
            pd.DataFrame: (See find)
        """
        #####
        def getNetworks(directory:str)->List[Network]:
            """
            Obtains the networks from a directory or serialization file.

            Args:
                directory (str): directory path or path to serialization file

            Returns:
                Networks
            """
            if directory.endswith(".txt"):
                serialization_path = directory
            else:
                # Construct the serialization file path and file
                serialization_path = os.path.join(directory, cn.SERIALIZATION_FILE)
            # Get the networks
                serializer = ModelSerializer(directory, serialization_path)
                if not os.path.exists(serialization_path):
                    serializer.serialize()
                collection = serializer.deserialize()
            return collection.networks
        #####
        reference_networks = getNetworks(reference_directory)
        target_networks = getNetworks(target_directory)
        # Put the serialized models in the directory. Check for it on invocation.
        finder = cls.makeFromCombinations(reference_networks, target_networks, identity=identity)
        return finder.find(is_report=is_report)


import pySubnetSB.constants as cn  # type: ignore
from pySubnetSB.network import Network # type: ignore

import numpy as np
import copy
import pandas as pd # type: ignore
import tellurium as te  # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
NUM_ITERATION = 2
NETWORK_NAME = "test"
BIG_NETWORK = """
J0: S1 -> S1 + S2; k1*S1;
J1: S2 -> S3; k2*S2;
J2: S3 -> ; k3*S3;
J3: S3 -> S4; k4*S3;
J4: S4 -> S5; k5*S4;

k1 = 1
k2 = 1
k3 = 1
k4 = 1
k5 = 1
S1 = 10
S2 = 0
S3 = 0
S4 = 0
S5 = 0
"""
NETWORK1 = """
J0: S1 -> S1 + S2; k1*S1;
J1: S2 -> S3; k2*S2;

k1 = 1
k2 = 1
S1 = 10
S2 = 0
S3 = 0
"""
NETWORK2 = """
// Strong structural identity with Network1
Ja: S2 -> S2 + S1; k1*S2;
Jb: S1 -> S3; k2*S1;

k1 = 3
k2 = 1
S1 = 1
S2 = 0
S3 = 0
"""
NETWORK3 = """
// Weak sructural identity with Network1
Ja: 2 S1 -> 2 S1 + S2; k1*S1*S1
Jb: S2 -> S3; k2*S1;

k1 = 3
k2 = 1
S1 = 1
S2 = 0
S3 = 0
"""
NETWORK4 = """
// Not structurally identical with Network1
Ja: S1 -> S2; k1*S2;
Jb: S1 -> S3; k2*S1;

k1 = 3
k2 = 1
S1 = 1
S2 = 0
S3 = 0
"""
NETWORK = Network.makeFromAntimonyStr(NETWORK1, network_name=NETWORK_NAME)


class TestNetwork(unittest.TestCase):

    def setUp(self):
        self.network = copy.deepcopy(NETWORK)

    def testConstrutor(self):
        if IGNORE_TEST:
            return
        self.assertEqual(self.network.network_name, NETWORK_NAME)
        self.assertTrue("int" in str(type(self.network.network_hash)))
        self.assertTrue("int" in str(type(self.network.network_hash)))

    def makeRandomNetwork(self, num_species=5, num_reaction=5):
        big_reactant_mat = np.random.randint(0, 2, (num_species, num_reaction))
        big_product_mat = np.random.randint(0, 2, (num_species, num_reaction))
        return Network(big_reactant_mat, big_product_mat)

    def testIsStructurallyIdenticalBasic(self):
        if IGNORE_TEST:
            return
        target, assignment_pair = self.network.permute()
        result = self.network.isStructurallyIdentical(target, identity=cn.ID_WEAK,
              is_report=False)
        self.assertTrue(np.all(self.network.species_names == target.species_names[assignment_pair.species_assignment]))
        self.assertTrue(np.all(self.network.reaction_names == target.reaction_names[assignment_pair.reaction_assignment]))
        self.assertTrue(result)
        result = self.network.isStructurallyIdentical(self.network, identity=cn.ID_STRONG,
              is_report=False)
        self.assertTrue(result)

    def testIsStructurallyIdenticalDoubleFail(self):
        if IGNORE_TEST:
            return
        # Double the size of the target. Look for exact match. Should fail.
        reactant_arr = np.vstack([self.network.reactant_nmat.values]*2)
        reactant_arr = np.hstack([reactant_arr]*2)
        product_arr = np.vstack([self.network.product_nmat.values]*2)
        product_arr = np.hstack([product_arr]*2)
        target_network = Network(reactant_arr, product_arr)
        result = self.network.isStructurallyIdentical(target_network, is_subnet=False, identity=cn.ID_WEAK)
        self.assertFalse(result)
    
    def testIsStructurallyIdenticalDoubleSubset(self):
        if IGNORE_TEST:
            return
        reactant_arr = np.hstack([self.network.reactant_nmat.values]*2)
        product_arr = np.hstack([self.network.product_nmat.values]*2)
        target_network = Network(reactant_arr, product_arr)
        result = self.network.isStructurallyIdentical(target_network, is_subnet=True, identity=cn.ID_WEAK)
        self.assertTrue(result)

    def testIsStructurallyIdenticalSimpleRandomlyPermute(self):
        if IGNORE_TEST:
            return
        target, _ = self.network.permute()
        result = self.network.isStructurallyIdentical(target, identity=cn.ID_WEAK)
        self.assertTrue(result)
        result = self.network.isStructurallyIdentical(target, identity=cn.ID_STRONG)
        self.assertTrue(result)

    def checkEquivalent(self, network1:Network, network2:Network=None, identity:str=cn.ID_STRONG):
        if network2 is None:
            network2 = self.network
        result = network1.isStructurallyIdentical(network2, identity=identity)
        self.assertTrue(result)
        result = network2.isStructurallyIdentical(network1, identity=identity)
        self.assertTrue(result)
    
    def testIsStructurallyIdenticalScaleRandomlyPermuteTrue(self):
        if IGNORE_TEST:
            return
        def test(reference_size, fill_factor=1, num_iteration=NUM_ITERATION):
            for identity in cn.ID_LST:
                for _ in range(num_iteration):
                    reference = Network.makeRandomNetworkByReactionType(num_reaction=reference_size,
                                                                        num_species=14)
                    target = reference.fill(num_fill_reaction=fill_factor*reference_size,
                        num_fill_species=fill_factor*reference_size)
                    result = reference.isStructurallyIdentical(target, identity=identity, is_subnet=True,
                          is_report=True)
                    msg = f"identity: {identity}, reference_size: {reference_size}, fill_factor: {fill_factor}"
                    msg += f"\n   num_species_candidate: {result.num_species_candidate}"
                    msg += f"\n   num_reaction_candidate: {result.num_reaction_candidate}"
                    #print(msg)
                    self.assertTrue(bool(result) or result.is_truncated)
        #
        for fill_factor in [1, 4]:
            for size in [5]:
                test(size, fill_factor=fill_factor)

    def testIsStructurallyIdenticalScaleRandomlyPermuteFalse(self):
        if IGNORE_TEST:
            return
        def test(reference_size, target_factor=1, num_iteration=NUM_ITERATION):
            num_truncate = 0
            num_processed = 0
            for _ in range(num_iteration):
                for identity in [cn.ID_WEAK, cn.ID_STRONG]:
                    for is_subnet in [False, True]:
                        reference = Network.makeRandomNetworkByReactionType(reference_size)
                        target = Network.makeRandomNetworkByReactionType(target_factor*reference_size)
                        # Analyze
                        result = reference.isStructurallyIdentical(target, identity=identity, is_subnet=is_subnet,
                            max_num_assignment=100)
                        num_truncate += result.is_truncated
                        num_processed += 1
                        #self.assertFalse(result)
            #print(f"Truncate rate: {num_truncate/num_processed}")
        #
        for size in [5, 10, 20]:
            test(size, target_factor=3)

    def testSerializeDeserialize(self):
        if IGNORE_TEST:
            return
        for _ in range(10):
            network = self.makeRandomNetwork(10, 10)
            serialization_str = network.serialize()
            self.assertTrue(isinstance(serialization_str, str))
            new_network = Network.deserialize(serialization_str)
            self.assertEqual(network, new_network)
    
    def testIsIsomorphic(self):
        if IGNORE_TEST:
            return
        def test(reference_size, is_isomorphic=True, num_iteration=NUM_ITERATION):
            for _ in range(num_iteration):
                    reference = Network.makeRandomNetworkByReactionType(reference_size)
                    target, _ = reference.permute()
                    if not is_isomorphic:
                        if target.reactant_nmat.values[0, 0] == 1:
                            target.reactant_nmat.values[0, 0] = 0
                        else:
                            target.reactant_nmat.values[0, 0] = 1
                    target = Network(target.reactant_nmat.values, target.product_nmat.values)
                    result = reference.isIsomorphic(target)
                    self.assertEqual(result, is_isomorphic)
        #
        test(10, is_isomorphic=True)
        test(10, is_isomorphic=False)
    
    def testIsStructurallyIdenticalBug(self):
        if IGNORE_TEST:
            return
        reference_mdl = """
        JJ0: S3 -> S2; 1
        JJ1: S2 -> S2; 1
        JJ2: S2 -> S4; 1
        JJ3: S7 -> S2 + S7; 1
        S1=0;
        S2=0;
        S3=0;
        S4=0;
        S5=0;
        S6=0;
        S7=0;
        S0=0;
        """
        target_mdl = """ 
        JJ2: S2 -> S4;1
        J0:  -> S1;1
        JJ3: S7 -> S7 + S2;1
        J3: S1 -> S6;1
        JJ1: S2 -> S2;1
        J2: S3 + S0 -> S1 + S7 + S4;1
        J1: S5 -> S2;1
        JJ0: S3 -> S2;1
        S1=0;
        S2=0;
        S3=0;
        S4=0;
        S5=0;
        S6=0;
        S7=0;
        S0=0;
        """
        reference_network = Network.makeFromAntimonyStr(reference_mdl)
        target_network = Network.makeFromAntimonyStr(target_mdl)
        result = reference_network.isStructurallyIdentical(target_network, is_subnet=True, num_process=1)
        self.assertTrue(result)


if __name__ == '__main__':
    unittest.main(failfast=False)


from pySubnetSB.assignment_pair import AssignmentPair  # type: ignore
import pySubnetSB.constants as cn  # type: ignore

import copy
import numpy as np
import unittest


IGNORE_TEST = False
IS_PLOT = False
SPECIES_ASSIGNMENT = np.array(range(3))
REACTION_ASSIGNMENT = np.array(range(4))


#############################
# Tests
#############################
class TestAssignmentPair(unittest.TestCase):

    def setUp(self):
        self.assignment_pair = AssignmentPair(species_assignment=SPECIES_ASSIGNMENT, reaction_assignment=REACTION_ASSIGNMENT)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(np.all(self.assignment_pair.species_assignment==SPECIES_ASSIGNMENT))
        self.assertTrue(np.all(self.assignment_pair.reaction_assignment==REACTION_ASSIGNMENT))

    def testEq(self):
        if IGNORE_TEST:
            return
        self.assertEqual(self.assignment_pair, self.assignment_pair)
        assignment_pair = AssignmentPair(species_assignment=REACTION_ASSIGNMENT, reaction_assignment=REACTION_ASSIGNMENT)
        self.assertNotEqual(self.assignment_pair, assignment_pair)

    def testSerializeDeserialize(self):
        if IGNORE_TEST:
            return
        serialization_str = self.assignment_pair.serialize()
        assignment_pair = self.assignment_pair.deserialize(serialization_str)
        self.assertEqual(self.assignment_pair, assignment_pair)

    def testInvert(self):
        if IGNORE_TEST:
            return
        def check(perm, inv_perm):
            identity_perm = range(len(perm))
            self.assertTrue(np.all(perm[inv_perm] == identity_perm))
        #####
        assignment_pair = self.assignment_pair.invert()
        check(self.assignment_pair.species_assignment, assignment_pair.species_assignment)
        check(self.assignment_pair.reaction_assignment, assignment_pair.reaction_assignment)

    def testMakeNameDct(self):
        if IGNORE_TEST:
            return
        with self.assertRaises(ValueError):
            self.assignment_pair.makeNameDct()
        #
        reference_species_names = ["species_" + str(n) for n in range(len(SPECIES_ASSIGNMENT))]
        reference_reaction_names = ["reaction_" + str(n) for n in range(len(REACTION_ASSIGNMENT))]
        target_species_names = ["species_" + str(n) for n in range(len(SPECIES_ASSIGNMENT)+1)]
        target_reaction_names = ["reaction_" + str(n) for n in range(len(REACTION_ASSIGNMENT)+1)]
        assignment_pair = AssignmentPair(species_assignment=SPECIES_ASSIGNMENT, reaction_assignment=REACTION_ASSIGNMENT,
              reference_reaction_names=reference_reaction_names, reference_species_names=reference_species_names,
              target_reaction_names=target_reaction_names, target_species_names=target_species_names)
        dct = assignment_pair.makeNameDct()
        self.assertTrue(cn.SPECIES_NAMES in dct)
        self.assertTrue(cn.REACTION_NAMES in dct)
        for superkey in [cn.SPECIES_NAMES, cn.REACTION_NAMES]:
            for key, value in dct[superkey].items():
                self.assertEqual(key, value)

        

if __name__ == '__main__':
    unittest.main()


'''Serizalizes Antimony and SBML models as Networks. Can run as a main program.'''

import pySubnetSB.constants as cn  # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.network_collection import NetworkCollection  # type: ignore

import os
import pandas as pd # type: ignore
import argparse
from typing import Optional, List

DEFAULT_SERIALIZATION_FILENAME = "network_serializers.txt"


BATCHSIZE = 20

class ModelSerializer(object):

    def __init__(self, model_directory:Optional[str]=None,
          serialization_path:str=DEFAULT_SERIALIZATION_FILENAME)->None:
        """
        The constructor has two forms:
        1. A model directory is given and optionally a name is given for the serialization file.
        2. A serialization path is given and the model directory is None.
        Args:
            model_directory (str): Path to directory that contains the model files
            serialization_file (str): Path for file where serialization results are stored
        """
        self.model_directory = model_directory
        self.serialization_file = serialization_path
        # self.serialization_path is the path to the serialization file
        if model_directory is None:
            if serialization_path.endswith("txt"):
                self.serialization_path = serialization_path
            else:
                raise ValueError("model_directory should be set if serialization_file is not a txt file.")
        else:
            if "/" in serialization_path:
                self.serialization_path = serialization_path
            else:
                self.serialization_path = os.path.join(model_directory, serialization_path)

    def remove(self)->None:
        """Removes the serialization file."""
        if os.path.exists(self.serialization_path):
            os.remove(self.serialization_path)

    @classmethod
    def serializerFromNetworks(cls, networks:List[Network], serialization_file:str,
          is_initialize:bool=False)->None:
        """
        Creates a serializer from a list of networks.

        Args:
            networks (List[Network]): List of networks
            serialization_file (str): Path to serialization file
            is_initialize (bool): If True, initializes the serialization file
        """
        serializer = cls(None, serialization_file)
        serializer.serializeNetworks(networks, is_initialize=is_initialize)

    @classmethod
    def makeOscillatorSerializer(cls, oscillator_directory:str,
          parent_directory:str=cn.OSCILLATOR_PROJECT)->'ModelSerializer':
        """
        Creates a serializer for the oscillators.

        Args:
            oscillator_directory (str): Name of oscillator directory
            parent_directory (str): 

        Returns:
            ModelSerializer: _description_
        """
        model_directory = os.path.join(parent_directory, oscillator_directory)
        return cls(model_directory, DEFAULT_SERIALIZATION_FILENAME)

    def serialize(self, batch_size:int=BATCHSIZE, num_batch:Optional[int]=None,
                           report_interval:Optional[int]=10,
                           serialization_strs:Optional[List[str]]=None)->None:
        """
        Serializes Antimony models in a directory.

        Args:
            batch_size (int): Number of models to process in a batch
            num_batch (Optional[int]): Number of batches to process
            report_interval (int): Interval to report progress. 
        """
        # Check if there is an existing output file
        processed_network_names:list = []
        if serialization_strs is None:
            if os.path.exists(self.serialization_path):
                with open(self.serialization_path, 'r') as f:
                    serialization_strs = f.readlines()
            else:
                serialization_strs = []
        processed_network_names = []
        for serialization_str in serialization_strs:
            network = Network.deserialize(serialization_str)
            processed_network_names.append(network.network_name)
        batch_count = 0
        while True:
            batch_count += 1
            if num_batch is not None and batch_count > num_batch:
                break
            #
            network_collection = NetworkCollection.makeFromAntimonyDirectory(self.model_directory,
                batch_size=batch_size,
                processed_network_names=processed_network_names, report_interval=report_interval)
            if len(network_collection) == 0:    
                break
            with open(self.serialization_path, 'a') as f:
                for network in network_collection.networks:
                    processed_network_names.append(network.network_name)
                    try:
                        stg = f'{network.serialize()}\n'
                        f.write(stg)
                    except:
                        print(f"**Error serializing {network.network_name}")
        if report_interval is not None:
            print("Done!")
    
    def serializeNetworks(self, networks:List[Network], is_initialize:bool=False)->None:
        """
        Serializes Networks
        """
        if is_initialize:
            mode = 'w'
        else:
            mode = 'a'
        with open(self.serialization_path, mode) as f:
            for network in networks:
                f.write(f'{network.serialize()}\n')

    @classmethod
    def deserializeFromStrings(cls, strings:List[str],
          model_names:Optional[List[str]]=None)->List[Network]:
        """Deserializes the network collection from a list of strings.
        Args:
            List of strings
            model_names (List[str]): If set, only networks with these names are deserialized.
        Returns:
            List[Network]: List of networks
        """
        networks = []
        for serialization_str in strings:
            if len(serialization_str) == 0:
                continue
            network = Network.deserialize(serialization_str)
            if network.num_species > 0:
                if (model_names is None) or (network.network_name in model_names):
                    networks.append(network)
        return networks

    def deserialize(self, model_names:Optional[List[str]]=None)->NetworkCollection:
        """Deserializes the network collection.
        Args:
            model_names (List[str]): If set, only networks with these names are deserialized.
        Returns:
            NetworkCollection: Collection of networks
        """
        with open(self.serialization_path, 'r') as f:
            serialization_strs = f.readlines()
        networks = self.deserializeFromStrings(serialization_strs, model_names=model_names)
        return NetworkCollection(networks, directory=self.model_directory)

# Run as a main program
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Serialize Antimony Models')
    parser.add_argument('model_directory', type=str, help='Name of directory')
    args = parser.parse_args()
    serializer = ModelSerializer(args.model_directory)
    serializer.serialize(report_interval=BATCHSIZE)


'''Calculates the significance level of finding an induced subnetwork in a target network.'''

"""
The null distribution are randomly generated models tuned to the BioModels repository.
A monte carlo simulation is used to generate the null distribution, and test models to see if
a reference network is present in the target network.

TODO:
1. Automatically choose target sizes to evaluate.
"""

import pySubnetSB.constants as cn # type: ignore
from pySubnetSB.significance_calculator_core import  SignificanceCalculatorCore, SignificanceCalculatorCoreResult # type: ignore
from pySubnetSB.network import Network  # type: ignore

import collections
import matplotlib.pyplot as plt
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
import seaborn as sns # type: ignore
from typing import List, Optional, Tuple  # type: ignore
import tqdm # type: ignore


NUM_REFERENCE_SPECIES = "num_reference_species"
NUM_REFERENCE_REACTION = "num_reference_reaction"
NUM_TARGET_SPECIES = "num_target_species"
NUM_TARGET_REACTION = "num_target_reaction"
FRAC_INDUCED = "frac_induced"
FRAC_TRUNCATED = "frac_truncated"
# Default values
NUM_ITERATION = 1000


PlotSignificanceResult = collections.namedtuple("PlotSignificanceResult",
      ["target_sizes", "frac_induces", "frac_truncates"])   


class SignificanceCalculator(object):

    def __init__(self, reference_network:Network, num_target_species:int, num_target_reaction:int,
                 identity:str=cn.ID_WEAK)->None:
        self.reference_network = reference_network
        self.num_target_reaction = num_target_reaction
        self.num_target_species = num_target_species
        self.identity = identity
    
    def calculate(self, num_iteration:int, is_report:bool=True,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT)->SignificanceCalculatorCoreResult:
        """
        Calculates the significance level of finding an induced subnetwork in a target network.

        Args:
            num_iteration (int): Number of iterations
            is_report (bool): If True, report progress
            max_num_assignment (int): Maximum number of assignment pairs

        Returns:
            SignificanceCalculatorResult
        """
        core_calculator = SignificanceCalculatorCore(self.num_target_species, self.num_target_reaction,
              num_target_network=num_iteration)
        return core_calculator.calculateSubnet(self.reference_network, identity=self.identity,
                max_num_assignment=max_num_assignment, is_report=is_report)
    
    @classmethod
    def generateNullDistribution(cls, num_iteration:int=1000, identity:str=cn.ID_WEAK,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
          reference_sizes:List[int]=[2*n for n in range(1, 11)],
          target_sizes:List[int]=[5*n + 20 for n in range(18)])->pd.DataFrame:
        """
        Generates a null distribution of finding an induced subnetwork in a target network.

        Args:
            reference_network (Network): Reference network
            num_target_reaction (int): Number of target species
            num_target_reaction (int): Number of target reactions
            num_iteration (int): Number of iterations
            identity (str): Identity type
            max_num_assignment (int): Maximum number of assignment pairs

        Returns:
            pd.DataFrame: Columns
                reference_size, target_size, frac_induced, frac_truncated
        """
        REFERENCE_SIZE = "reference_size"
        TARGET_SIZE = "target_size"
        dct:dict = {n: [] for n in [REFERENCE_SIZE, TARGET_SIZE, FRAC_INDUCED, FRAC_TRUNCATED]}
        for reference_size in reference_sizes:
            for target_size in target_sizes:
                reference_network = Network.makeRandom(reference_size, reference_size)
                calculator = SignificanceCalculator(reference_network, target_size, target_size,
                      identity=identity)
                result = calculator.calculate(num_iteration, max_num_assignment=max_num_assignment)
                dct[REFERENCE_SIZE].append(reference_size)
                dct[TARGET_SIZE].append(target_size)
                dct[FRAC_INDUCED].append(result.frac_induced)
                dct[FRAC_TRUNCATED].append(result.frac_truncated)
        # Create the DataFrame
        df = pd.DataFrame(dct)
        return df

    def plotSignificance(self, target_sizes:Optional[List[int]]=None, num_iteration:int=NUM_ITERATION,
          max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
          is_report:bool=True, is_plot=True)->PlotSignificanceResult:
        """
        Plots fraction of random targets in which the reference is used.
        1 - fraction_induced is significance level.

        Args:
            target_sizes (List[int]): _description_
            num_iteration (int, optional): _description_. Defaults to 1000.
            is_report (bool, optional): Report progress. Defaults to True.
            is_plot (bool, optional): Plot

        Returns:
            PlotSignificanceResult

        To further adorn the plot, use is_plot=False and then plt.gca() to get the axis.
        """
        # Set default target sizes
        if target_sizes is None:
            min_size = max(self.reference_network.num_species, self.reference_network.num_reaction)
            target_sizes = [n for n in range(min_size, 6*min_size, min_size)]
        frac_induces = []
        frac_truncates = []
        for target_size in tqdm.tqdm(target_sizes, desc="targets", disable=not is_report):
            calculator = SignificanceCalculator(self.reference_network, target_size,
                target_size, identity=self.identity)
            result = calculator.calculate(num_iteration, max_num_assignment=max_num_assignment,
                is_report=False)
            frac_induces.append(result.frac_induced)
            frac_truncates.append(result.frac_truncated)
        _ = plt.bar(target_sizes, frac_induces, color='blue', alpha=0.5)
        _ = plt.bar(target_sizes, frac_truncates, bottom=frac_induces, color='red', alpha=0.5)
        plt.ylim([0, 1])
        plt.xlabel("Target size")
        plt.ylabel("fraction induced")
        plt.legend(["induced", "truncated"])
        if is_plot:
            plt.show()
        return PlotSignificanceResult(target_sizes=target_sizes, frac_induces=frac_induces,
            frac_truncates=frac_truncates)

    @classmethod 
    def calculateNetworkOccurrenceProbability(cls, reference_network:Network, num_iteration:int=10000,
              identity:str=cn.ID_STRONG, is_report:bool=True,
              max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT)->Tuple[float, float]:
        """
        Calculates the probability of finding an induced subnetwork in a target network.

        Args:
            reference_network (Network): Reference network
            num_iteration (int): Number of iterations

        Returns:
            float: fraction_induced, fraction_truncated
        """
        """ calculator = SignificanceCalculator(reference_network, reference_network.num_species,
              reference_network.num_reaction)
        result = calculator.calculate(num_iteration, is_report=False)
        return result.frac_induced, result.frac_truncated """
        initial_networks = [Network.makeRandomNetworkByReactionType(reference_network.num_reaction,
              reference_network.num_species, is_exact=True)
             for _ in range(num_iteration)]
        networks = [n for n in initial_networks if n.network_hash == reference_network.network_hash]
        # Calculate the fraction of identical networks for those with the same hash
        num_identical = 0
        num_truncated = 0
        for network in networks:
            result = reference_network.isStructurallyIdentical(network, identity=identity,
                max_num_assignment=max_num_assignment,
                is_subnet=False, is_report=is_report)
            num_identical += 1 if result else 0
            num_truncated += 1 if result.is_truncated else 0
        return num_identical/num_iteration, num_truncated/num_iteration

    @classmethod 
    def plotProbabilityOfOccurrence(cls, data_df:pd.DataFrame, column:str,
          is_plot:bool=True)->None:
        """
        Plots mean and standard deviation (in parens) of the probability of occurrence
        of reaction networks in -log10 units.
          x - number of species
          y - number of reactions
          heatmap - probability of occurrence

        Args:
            data_df (pd.DataFrame): Dataframe with columns num_species
            column (str): Column to analyze. Must be a probability.
        """
        #####
        def transform(df:pd.DataFrame)->pd.DataFrame:
            df.sort_index(level=0, ascending=False, inplace=True)
            df.style.format({c: '{:.2f}' for c in df.columns})
            return df
        #####
        # Transform the data
        df = data_df.copy()
        df[column] = df[column].map(lambda x: max(x, 1e-5))
        df[column] = df[column].map(lambda x: -np.log10(x))
        mean_df = df.pivot_table(index=cn.D_NUM_SPECIES, columns=cn.D_NUM_REACTION,
              values=column, aggfunc='median')
        mean_df = transform(mean_df)
        std_df = df.pivot_table(index=cn.D_NUM_SPECIES, columns=cn.D_NUM_REACTION,
              values=column, aggfunc='std')
        std_df = transform(std_df)
        # Create the labels
        label_df = mean_df.copy().astype(str)
        for index in mean_df.index:
            for column in mean_df.columns:
                is_assigned = False
                if not np.isnan(mean_df.loc[index, column]):
                    if index in std_df.index and column in std_df.columns:
                        text = f"{mean_df.loc[index, column]:1.2f}\n({std_df.loc[index, column]:1.2f})"
                        text = text.replace("(nan)", "")
                        label_df.loc[index, column] = text
                        is_assigned = True
                if not is_assigned:
                    label_df.loc[index, column] = ""
        ax = sns.heatmap(mean_df, annot=label_df.values, fmt="", cmap="coolwarm", vmin=0, vmax=5,
              annot_kws={'size': 8},
              cbar_kws={'label': '-log10 Probability of occurrence'})
        if is_plot:
            plt.show()


from pySubnetSB import constants as cn  # type: ignore
from pySubnetSB.network import Network # type: ignore
from pySubnetSB.matrix import Matrix # type: ignore
from pySubnetSB.network_collection import NetworkCollection # type: ignore
from pySubnetSB.processed_network import ProcessedNetwork # type: ignore
from pySubnetSB.cluster_builder import ClusterBuilder # type: ignore

import copy
import pandas as pd # type: ignore
import numpy as np # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
COLLECTION_SIZE = 10
if not IGNORE_TEST:
    NETWORK_COLLECTION = NetworkCollection.makeRandomCollection(num_network=COLLECTION_SIZE)


#############################
# Tests
#############################
class TestClusterBuilder(unittest.TestCase):

    def setUp(self):
        if IGNORE_TEST:
            return
        network_collection = copy.deepcopy(NETWORK_COLLECTION)
        self.builder = ClusterBuilder(network_collection, is_report=IS_PLOT,
                                      max_num_assignment=100, identity=cn.ID_WEAK)

    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertTrue(len(self.builder.network_collection) == COLLECTION_SIZE)
        self.assertTrue(isinstance(self.builder.network_collection, NetworkCollection))

    def testMakeHashDct(self):
        if IGNORE_TEST:
            return
        hash_dct = self.builder._makeHashDct()
        count = np.sum([len(v) for v in hash_dct.values()])  
        self.assertTrue(count == COLLECTION_SIZE)
        #
        for networks in hash_dct.values():
            for network in networks:
                is_true = any([network == n for n in self.builder.network_collection.networks])
                self.assertTrue(is_true)

    def makeStructurallyIdenticalCollection(self, num_network:int=5, num_species:int=5, num_reaction:int=7):
        """Makes a structurally identical collection with strong identity.

        Args:
            num_network (int, optional): _description_. Defaults to 5.
            num_row (int, optional): _description_. Defaults to 5.
            num_column (int, optional): _description_. Defaults to 7.

        Returns:
            _type_: _description_
        """
        """ array1 = np.random.randint(0, 3, size=(num_species, num_reaction))
        array2 = np.random.randint(0, 3, size=(num_species, num_reaction))
        network = Network(array1, array2) """
        network = Network.makeRandomNetworkByReactionType(num_species=num_species, num_reaction=num_reaction)
        networks = []
        for _ in range(num_network):
            new_network, _ = network.permute()
            networks.append(new_network)
        return NetworkCollection(networks)

    def makeStronglyIdenticalCollection(self, num_network:int=5, num_row:int=5, num_column:int=7):
        array1 = Matrix.makeTrinaryMatrix(num_row=num_row, num_column=num_column)
        array2 = Matrix.makeTrinaryMatrix(num_row=num_row, num_column=num_column)
        network = Network(array1, array2)
        networks = []
        for _ in range(num_network):
            new_network, _ = network.permute()
            networks.append(new_network)
        return NetworkCollection(networks)

    def testCluster(self):
        if IGNORE_TEST:
            return
        # Construct a collection of two sets of permutably identical matrices
        def test(num_collection=2, num_network=5, num_species=15, num_reaction=15,
                 identity=cn.ID_STRONG):
            is_success = True
            for _ in range(5):  # Do multiple times to handle edge cases
                # Make disjoint network collections, each of which is structurally identical
                network_collections = [self.makeStructurallyIdenticalCollection(
                    num_species=num_species, num_reaction=num_reaction, num_network=num_network)
                    for _ in range(num_collection)]
                # Construct the network_collection to analyze that it is the combination of the other network_collections
                network_collection = network_collections[0]
                for network in network_collections[1:]:
                    try:
                        network_collection += network
                    except ValueError:
                        # Duplicate randomly generated name. Ignore.
                        pass
                #
                builder = ClusterBuilder(network_collection, max_num_assignment=100000, is_report=IS_PLOT,
                                        identity=identity)
                builder.cluster()
                num_builder_collection = len(builder.processed_network_collections)
                if num_builder_collection >  num_collection:
                    num_indeterminate = np.sum([b.in_determinate for b in builder.processed_network_collections])
                    if num_indeterminate == num_builder_collection - num_collection:
                        is_success = True
                else:
                    if num_builder_collection == num_collection:
                        is_success = True
                for network_collection in network_collections:
                    self.assertTrue(str(network_collection) in str(network_collections))
                if is_success:
                    break
            self.assertTrue(is_success)
        #
        #test(num_collection=5, num_network=1000, num_species=15, num_reaction=15)
        test(num_species=4, num_reaction=4)
        test(num_collection=5, num_network=10)
        test(num_collection=5)
        test(num_collection=15, identity=cn.ID_WEAK)
            
    def testMakeNetworkCollection(self):
        if IGNORE_TEST:
            return
        ARRAY_SIZE = 5
        network_collection = NetworkCollection.makeRandomCollection(num_species=ARRAY_SIZE,
              num_reaction=ARRAY_SIZE, num_network=COLLECTION_SIZE)
        processed_networks = [ProcessedNetwork(network) for network in network_collection.networks]
        builder = ClusterBuilder(network_collection, is_report=IS_PLOT)
        for idx, processed_network in enumerate(processed_networks):
            network = builder.makeNetworkFromProcessedNetwork(processed_network)
            self.assertTrue(network == network_collection.networks[idx])

if __name__ == '__main__':
    unittest.main()


import pySubnetSB.constants as cn  # type: ignore
import pySubnetSB.parallel_subnet_finder_worker as psfw  # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.mock_queue import MockQueue  # type: ignore
from pySubnetSB.model_serializer import ModelSerializer  # type: ignore
from pySubnetSB.parallel_subnet_finder_worker import SubnetFinderWorkunitManager  # type: ignore
from pySubnetSB.worker_checkpoint_manager import WorkerCheckpointManager  # type: ignore

import json
import os
import pandas as pd # type: ignore
import numpy as np
from typing import Tuple
import unittest

IGNORE_TEST = False
IS_PLOT =  False
SIZE = 10
NUM_NETWORK = 10
TOTAL_WORKER = 2
BASE_CHECKPOINT_PATH = os.path.join(cn.TEST_DIR, "test_parallel_subnet_finder_worker_checkpoint.csv")
WORKUNIT_PATH = os.path.join(cn.TEST_DIR, "test_parallel_subnet_finder_worker_manager.csv")
TASK0_CHECKPOINT_PATH = psfw.WorkerCheckpointManager.makeWorkerCheckpointPath(BASE_CHECKPOINT_PATH, 0)
TASK1_CHECKPOINT_PATH = psfw.WorkerCheckpointManager.makeWorkerCheckpointPath(BASE_CHECKPOINT_PATH, 1)
TASK_CHECKPOINT_PATHS = [TASK0_CHECKPOINT_PATH, TASK1_CHECKPOINT_PATH]
REFERENCE_SERIALIZER_PATH = os.path.join(cn.TEST_DIR,
      "test_parallel_subnet_finder_worker_references.txt")
TARGET_SERIALIZER_PATH = os.path.join(cn.TEST_DIR,
      "test_parallel_subnet_finder_worker_targets.txt")
REMOVE_FILES:list = [BASE_CHECKPOINT_PATH, TASK0_CHECKPOINT_PATH, TASK1_CHECKPOINT_PATH,
      REFERENCE_SERIALIZER_PATH, TARGET_SERIALIZER_PATH, WORKUNIT_PATH]
REFERENCE_NETWORKS = [Network.makeRandomNetworkByReactionType(SIZE, is_prune_species=True)
      for _ in range(NUM_NETWORK)]
TARGET_NETWORKS = [n.fill(num_fill_reaction=SIZE, num_fill_species=SIZE) for n in REFERENCE_NETWORKS]


#############################
# Tests
#############################
def makeDataframe(num_network:int)->pd.DataFrame:
    # Creates a dataframe used by the WorkerCheckpointManager
    reference_networks = [Network.makeRandomNetworkByReactionType(3, is_prune_species=True) for _ in range(num_network)]
    target_networks = [Network.makeRandomNetworkByReactionType(3, is_prune_species=True) for _ in range(num_network)]
    dct = {cn.FINDER_REFERENCE_NETWORK: [str(n) for n in range(num_network)],
           cn.FINDER_INDUCED_NETWORK: [str(n) for n in range(num_network, 2*num_network)]}
    df = pd.DataFrame(dct)
    df[cn.FINDER_REFERENCE_NAME] = [str(n) for n in reference_networks]
    df[cn.FINDER_TARGET_NAME] = [str(n) for n in target_networks]
    df[cn.FINDER_NAME_DCT] = [json.dumps(dict(a=n)) for n in range(num_network)]
    return df


#############################
class TestParallelSubnetFinderWorker(unittest.TestCase):

    def setUp(self):
        self.initializeNetworks()
        self.initializeWorkunitManager()

    def tearDown(self):
        self.remove()

    def remove(self):
        if hasattr(self, "manager"):
            self.manager.remove()
        for ffile in REMOVE_FILES:
            if os.path.exists(ffile):
                os.remove(ffile)

    def initializeNetworks(self, num_network:int=NUM_NETWORK):
        reference_networks = [Network.makeRandomNetworkByReactionType(SIZE, is_prune_species=True)
                  for _ in range(num_network)]
        target_networks = [n.fill(num_fill_reaction=SIZE, num_fill_species=SIZE) for n in reference_networks]
        ModelSerializer.serializerFromNetworks(reference_networks, REFERENCE_SERIALIZER_PATH,
                is_initialize=True)
        ModelSerializer.serializerFromNetworks(target_networks, TARGET_SERIALIZER_PATH,
                is_initialize=True)

    def initializeWorkunitManager(self, total_worker:int=TOTAL_WORKER):
        if hasattr(self, "manager"):
            self.manager.remove()  # type: ignore
        self.manager = SubnetFinderWorkunitManager.makeFromSerializationFiles(WORKUNIT_PATH,
              num_worker=total_worker, reference_serialization_path=REFERENCE_SERIALIZER_PATH,
              target_serialization_path=TARGET_SERIALIZER_PATH)
        self.manager.makeWorkunitFile()

    def testExecuteTask(self):
        #if IGNORE_TEST:
        #    return
        #####
        def test(task_idx:int=0, is_initialize:bool=True, total_worker:int=2)->Tuple[pd.DataFrame, pd.DataFrame]:
            # Creates reference and target networks to assess processing by tasks.
            # Returns the full dataframe and the pruned dataframes
            psfw.executeWorker(task_idx, WORKUNIT_PATH, total_worker, BASE_CHECKPOINT_PATH,
                REFERENCE_SERIALIZER_PATH, TARGET_SERIALIZER_PATH, identity=cn.ID_STRONG,
                is_report=IS_PLOT, is_initialize=is_initialize, is_allow_exit=False)
            checkpoint_manager = psfw.WorkerCheckpointManager(TASK_CHECKPOINT_PATHS[task_idx],
                    is_report=IS_PLOT, is_initialize=False)
            full_df = checkpoint_manager.recover().full_df
            if len(full_df) == 0:
                return full_df, full_df
            prune_result = psfw.WorkerCheckpointManager.prune(full_df)
            return prune_result.full_df, prune_result.pruned_df
        #####
        # Finds the subnets for a single task
        total_worker = 1
        self.initializeWorkunitManager(total_worker=total_worker)
        full_df, prune_df = test(task_idx=0, total_worker=total_worker, is_initialize=True)
        self.assertEqual(len(full_df), NUM_NETWORK**2/total_worker)
        #self.assertLessEqual(np.abs(len(prune_df) - NUM_NETWORK/total_worker), 1)
        # Finds all subnets if there is an existing checkpoint
        # FIXME: Test for recovering a checkpoint
        total_worker = 1
        self.initializeWorkunitManager(total_worker=total_worker)
        _ = test(task_idx=0, is_initialize=True, total_worker=total_worker)
        full2_df, _ = test(task_idx=0, is_initialize=False, total_worker=total_worker)
        self.assertEqual(len(full2_df), NUM_NETWORK**2)
        # Handle two tasks
        total_worker = 2
        WorkerCheckpointManager.deleteWorkerCheckpoints(WORKUNIT_PATH, total_worker)
        self.initializeWorkunitManager(total_worker=total_worker)
        for task_idx in range(2):
            full_df, prune_df = test(task_idx=task_idx, total_worker=total_worker,
                  is_initialize=True)
            self.assertEqual(len(full_df), NUM_NETWORK**2/2)


if __name__ == '__main__':
    unittest.main(failfast=False)


'''Compares the reference matrix with selected rows and columns in a target matrix.'''

"""
    Compares the reference matrix with selected rows and columns in a target matrix. The comparison is
    vectorized and checks if the reference matrix is identical to the target matrix when the rows and columns
    are permuted according to the selected assignments. The comparison is for equality, numerical inequality,
    or bitwise inequality.

    The comparison is performed by constructing a large matrix that is a concatenation of the reference matrix
    and the target matrix. The large matrix is constructed by repeating the reference matrix for each assignment
    of target rows to reference rows and target columns to reference columns. The target matrix is re-arranged
    according to the assignments. The large matrix is compared to the re-arranged target matrix. The comparison
    is performed for each assignment pair. The results are used to determine if the assignment pair results in
    an identical matrix.

    Some core concepts:
    * Assignment is a selection of a subset of rows (columns) in the target matrix to be compared to the reference matrix.
    * AssignmentPair is a pair of assignments of rows and columns in the target matrix to the reference matrix.
    * AssignmentPairIndex. An index into the cross-product of the row and column assignments. This is referred to
        as linear addressing. Using the separate indices for rows and columns is referred to as vector addressing.
        Methods are provided convert
        between the assignment pair index and the indices of row and column assignments.
    * AssignmentArray is a two dimensional array of assignments. Columns are the index of the target row that
        is assigned to the reference row (column index).
        Rows are instances of an assignment. There is an AssignmentArray for the rows and for the columns.
    * A Comparison is a comparison betweeen the reference matrix and an assignment of rows and columns in the target

 """
from pySubnetSB.assignment_pair import AssignmentPair # type: ignore
from pySubnetSB.assignment_evaluator_worker import AssignmentEvaluatorWorker, WorkerResult # type: ignore
from pySubnetSB import constants as cn # type: ignore
from pySubnetSB import util # type: ignore
from src.pySubnetSB.performance_monitor import PerformanceMonitor # type: ignore

from concurrent.futures import ProcessPoolExecutor
from memory_profiler import profile  # type: ignore
import numpy as np
import multiprocessing as mp
from typing import Union, Tuple, Optional, List


DEBUG = False
IS_ENABLE_PERFORMANCE_EVALUATION = False
MIN_ASSIGNMENT_PER_PARALLEL_CPU = int(1e2)



#########################################################
class AssignmentBatchContext(object):
    # Saves context for a batch of assignment evaluation
    def __init__(self, big_reference_arr, big_target_arr, big_target_row_idx_arr, big_target_column_idx_arr):
        self.big_reference_arr = big_reference_arr
        self.big_target_arr = big_target_arr
        self.big_target_row_idx_arr = big_target_row_idx_arr  # Indices for target rows
        self.big_target_column_idx_arr = big_target_column_idx_arr  # Indices for target columns


#########################################################
class AssignmentEvaluator(object):
       
    def __init__(self, reference_arr:np.ndarray, target_arr:np.ndarray, max_batch_size:int=cn.MAX_BATCH_SIZE):
        """
        Args:
            reference_arr (np.ndarray): reference matrix
            target_arr (np.ndarray): target matrix
            row_assignment_arr (np.ndarray): candidate assignments of target rows to reference rows
            column_assignment_arr (np.ndarray): candidate assignments of target columns to reference columns
            comparison_criteria (ComparisonCriteria): comparison criteria
            max_batch_size (int): maximum batch size in units of bytes
        """
        self.monitor = PerformanceMonitor("AssignmentEvaluator", is_enabled=IS_ENABLE_PERFORMANCE_EVALUATION)
        self.reference_arr = reference_arr
        self.num_reference_row, self.num_reference_column = self.reference_arr.shape
        self.target_arr = target_arr   # Reduce memory usage
        self.max_batch_size = max_batch_size

    def linearToVector(self, num_column_assignment:int,
          index:Union[int, np.ndarray])->Union[Tuple[int, int], Tuple[np.ndarray, np.ndarray]]:  
        """
        Converts a linear address of the candidate assignments to a vector index.
        Converts a vector representation of the candidate assignments (row, column) to a linear index

        Args:
            num_column_assignment (int): number of column assignments
            index (int): linear index

        Returns:
            row_idx (int): row index
            column_idx (int): column index
        """
        row_idx = index//num_column_assignment
        column_idx = index%num_column_assignment
        return row_idx, column_idx  # type: ignore

    #@profile
    def evaluateAssignmentPairs(self, assignment_pairs:List[AssignmentPair],
          max_num_assignment:int=-1)->List[AssignmentPair]:
        """Finds the pair of row and column assignments that satsify the comparison criteria.
        Uses np.isclose in comparisons

        Args:
            assignment_pairs: List[AssignmentPair] - pairs of row and column assignments
            max_num_assignment: int - maximum number of assignments (no bound if -1)

        Returns:
            List[AssignmentPair]: Assignment pairs that successfully compare
        """
        successful_assignment_pairs = []
        if max_num_assignment == -1:
            max_num_assignment = len(assignment_pairs)
        for assignment_pair in assignment_pairs:
            target_arr = self.target_arr[assignment_pair.row_assignment, :]
            target_arr = target_arr[:, assignment_pair.column_assignment]
            result = AssignmentEvaluatorWorker.compare(self.reference_arr, target_arr, self.num_reference_row,
                  is_close=True)
            if result:
                successful_assignment_pairs.append(assignment_pair)
            if len(successful_assignment_pairs) >= max_num_assignment:
                break
        return successful_assignment_pairs

    #@profile
    def parallelEvaluate(self, row_assignment_arr:np.ndarray, column_assignment_arr:np.ndarray,
                 total_process:int=-1, max_num_assignment:int=cn.MAX_NUM_ASSIGNMENT,
                 is_report:bool=True)->WorkerResult:
        """Evaluates the assignments for the target matrix using iteration.

        Args:
            row_assignment_arr (np.ndarray): assignments of target rows to reference rows
            column_assignment_arr (np.ndarray): assignments of target columns to reference columns
            num_process (int): number of processes
            total_process (int): number of processes (defaults to the number of CPUs)
            max_num_assignment (int): maximum number of assignments
            is_report (bool): report progress

        Returns:
            WorkerResult: results of the evaluation
                AssignmentPair: Row and column assignments that result in equality
                is_truncated: True if the evaluation was truncated
        """
        self.monitor.add("parallelEvaluate: Start")
        num_comparison = row_assignment_arr.shape[0]*column_assignment_arr.shape[0]
        # Initializations
        if total_process == -1:
            total_process = mp.cpu_count()
        num_process = min(mp.cpu_count(), total_process)
        if num_comparison < MIN_ASSIGNMENT_PER_PARALLEL_CPU:
            num_process = 1
        # Handle the case of a single process
        if num_process == 1:
            self.monitor.add("parallelEvaluate/Single Process: Start")
            # Construct the assignments
            row_assignments = [row_assignment_arr]
            column_assignments = [column_assignment_arr]
            # Run in a single process
            return_dct:dict = {}
            procnum = 0
            AssignmentEvaluatorWorker.do(self.reference_arr, self.target_arr, self.max_batch_size,
                  row_assignment_arr, column_assignment_arr, procnum, total_process,
                  return_dct, max_num_assignment,
                  is_report=is_report)
            self.monitor.add("parallelEvaluate/Single Process: End")
            return return_dct[0]
        else:
            self.monitor.add("parallelEvaluate/Multiple Process: Start")
            # Construct the assignments
            if row_assignment_arr.shape[0] > column_assignment_arr.shape[0]:
                row_assignments = util.partitionArray(row_assignment_arr, num_process)
                actual_num_process = len(row_assignments)
                column_assignments = [column_assignment_arr]
                is_row_max = True
            else:
                row_assignments = [row_assignment_arr]
                column_assignments = util.partitionArray(column_assignment_arr, num_process)
                actual_num_process = len(column_assignments)
                is_row_max = False
            self.monitor.add("parallelEvaluate/Multiple Process/Assignments: Start")
            # Construct the arguments
            args = []
            return_dct = {}
            for procnum in range(actual_num_process):
                if is_row_max:
                    process_row_assignment_arr = row_assignments[procnum]
                    process_column_assignment_arr = column_assignments[0]
                else:
                    process_row_assignment_arr = row_assignments[0]
                    process_column_assignment_arr = column_assignments[procnum]
                args.append((self.reference_arr, self.target_arr, self.max_batch_size,
                      process_row_assignment_arr, process_column_assignment_arr,
                      procnum, total_process, return_dct, max_num_assignment, is_report))
            self.monitor.add("parallelEvaluate/Multiple Process/Assignments: End")
            # Run the processes
            self.monitor.add("parallelEvaluate/Multiple Process/Run process pool: Start")
            with ProcessPoolExecutor(max_workers=total_process) as executor:
                process_args = zip(*args)
                results = executor.map(AssignmentEvaluatorWorker.do, *process_args)
            self.monitor.add("parallelEvaluate/Multiple Process/Run process pool: End")
        # Combine the results
        self.monitor.add("parallelEvaluate/Multiple Process/Combine results: Start")
        assignment_pairs = []
        is_truncated = False
        for result in results:
            is_truncated = is_truncated or result.is_truncated
            assignment_pairs.extend(result.assignment_pairs)
        worker_result = WorkerResult(assignment_pairs=assignment_pairs, is_truncated=is_truncated)
        self.monitor.add("parallelEvaluate/Multiple Process/Combine results: End")
        return worker_result
    
    def _manageEvaluation(self, reference_arr:np.ndarray, target_arr:np.ndarray, max_batch_size:int,
           row_assignment_arr:np.ndarray, column_assignment_arr:np.ndarray,
           procnum:int, total_proc:int)->List[AssignmentPair]:
        """Manages the evaluation of assignments.

        Args:
            reference_arr (np.ndarray): _description_
            target_arr (np.ndarray): _description_
            max_batch_size (int): _description_
            row_assignment_arr (np.ndarray): _description_
            column_assignment_arr (np.ndarray): _description_
            procnum (int): Number of this process
            total_proc (int): Number of processes

        Returns:
            List[AssignmentPair]: _description_
        """


        raise NotImplementedError("Subclass must implement this method")


import pySubnetSB.constants as cn  # type: ignore
from pySubnetSB.subnet_finder import SubnetFinder  # type: ignore
from pySubnetSB.network import Network  # type: ignore
from pySubnetSB.worker_checkpoint_manager import WorkerCheckpointManager # type: ignore

import json
import os
import pandas as pd # type: ignore
import numpy as np
import unittest

IGNORE_TEST = False
IS_PLOT =  False
SIZE = 3
MODEL_DIR = os.path.join(cn.TEST_DIR, "oscillators")
CHECKPOINT_PATH = os.path.join(cn.DATA_DIR, "test_subnet_finder_checkpoint.csv")


#############################
# Tests
#############################
def makeDataframe(num_network:int)->pd.DataFrame:
    # Creates a dataframe used by the CheckpointManager
    reference_networks = [Network.makeRandomNetworkByReactionType(3, is_prune_species=True) for _ in range(num_network)]
    target_networks = [Network.makeRandomNetworkByReactionType(3, is_prune_species=True) for _ in range(num_network)]
    dct = {cn.FINDER_REFERENCE_NETWORK: [str(n) for n in range(num_network)],
           cn.FINDER_INDUCED_NETWORK: [str(n) for n in range(num_network, 2*num_network)]}
    df = pd.DataFrame(dct)
    df[cn.FINDER_REFERENCE_NAME] = [str(n) for n in reference_networks]
    df[cn.FINDER_TARGET_NAME] = [str(n) for n in target_networks]
    df[cn.FINDER_NAME_DCT] = [json.dumps(dict(a=n)) for n in range(num_network)]
    return df


#############################
class TestSubnetFinder(unittest.TestCase):

    def setUp(self):
        self.reference = Network.makeRandomNetworkByReactionType(SIZE, is_prune_species=True)
        self.target = self.reference.fill(num_fill_reaction=SIZE, num_fill_species=SIZE)
        self.finder = SubnetFinder.makeFromCombinations(reference_networks=[self.reference], target_networks=[self.target],
              identity=cn.ID_WEAK)
        
    def testConstructor(self):
        if IGNORE_TEST:
            return
        self.assertGreater(len(self.finder.network_pairs), 0)

    def testFindSimple(self):
        if IGNORE_TEST:
            return
        df = self.finder.find(is_report=IS_PLOT)
        self.assertEqual(len(df), 1)

    def testFindScale(self):
        if IGNORE_TEST:
            return
        NUM_REFERENCE_MODEL = 10
        NUM_EXTRA_TARGET_MODEL = 10
        NETWORK_SIZE = 10
        fill_size = 3
        # Construct the models
        reference_models = [Network.makeRandomNetworkByReactionType(NETWORK_SIZE, is_prune_species=True)
              for _ in range(NUM_REFERENCE_MODEL)]
        target_models = [r.fill(num_fill_reaction=fill_size, num_fill_species=fill_size) for r in reference_models]
        # Add extra target models
        target_models += [Network.makeRandomNetworkByReactionType(NETWORK_SIZE, is_prune_species=True)
              for _ in range(NUM_EXTRA_TARGET_MODEL)]
        # Do the search
        finder = SubnetFinder.makeFromCombinations(reference_networks=reference_models, target_networks=target_models,
              identity=cn.ID_STRONG)
        df = finder.find(is_report=IS_PLOT)
        prune_df = WorkerCheckpointManager.prune(df).pruned_df
        self.assertEqual(len(prune_df), NUM_REFERENCE_MODEL)
    
    def testFindFromDirectories(self):
        if IGNORE_TEST:
            return
        df = SubnetFinder.findFromDirectories(MODEL_DIR, MODEL_DIR, identity=cn.ID_WEAK, is_report=IS_PLOT)
        num_unique = len(df[cn.FINDER_REFERENCE_NAME].unique())
        prune_df = WorkerCheckpointManager.prune(df).pruned_df
        num_match = np.sum(prune_df[cn.FINDER_REFERENCE_NAME] == prune_df[cn.FINDER_TARGET_NAME])
        self.assertTrue(num_match >= num_unique)


if __name__ == '__main__':
    unittest.main(failfast=True)


'''Timing for performance analysis'''

import os
import time
import psutil # type: ignore
from typing import List, Tuple


def getMemoryUsage():
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss # in bytes


class PerformanceMonitor(object):

    def __init__(self, name:str, is_enabled:bool=True):
        self.is_enabled = is_enabled
        self.name = name
        self.memory_usages:list = []
        self.elapsed_times:list = []
        self.base_time = time.time()
        self.names:list = [name + ": Start"]
        if self.is_enabled:
            print(f"\n**** {name}")

    def add(self, name:str):
        if not self.is_enabled:
            return
        self.names.append(name)
        elapsed_time = time.time() - self.base_time
        self.elapsed_times.append(elapsed_time)
        memory_usage = getMemoryUsage()/1e6  # in MB
        self.memory_usages.append(memory_usage)
        print(f"{self.names[-2]} - {name}: {elapsed_time}, {memory_usage}")
        self.base_time = time.time()


from pySubnetSB.stoichometry import Stoichiometry  # type: ignore
import pySubnetSB.constants as cn  # type: ignore

import os
import numpy as np
import tellurium as te # type: ignore
import unittest


IGNORE_TEST = False
IS_PLOT = False
NUM_MODELS = 10
MODEL_DIR = os.path.join(cn.TEST_DIR, "oscillators")
MODEL1 = """
model test
            J0: S1 -> S2; k1*S1;
            k1 = 0.1;
            S1 = 10;
        end
"""
MODEL2 = """
model test
            J0: $S1 -> S2; k1*S1;
            k1 = 0.1;
            S1 = 10;
        end
"""

#############################
# Tests
#############################
class TestStoichiometryMatrices(unittest.TestCase):

    def testSimple(self):
        if IGNORE_TEST:
            return
        antimony_str = MODEL2
        stoichiometry = Stoichiometry(antimony_str)
        self.assertTrue(np.all(stoichiometry.reactant_mat == np.array([[0], [1]])))
        self.assertTrue(np.all(stoichiometry.product_mat == np.array([[1], [0]])))
        self.assertTrue(np.all(stoichiometry.standard_mat == np.array([[1], [-1]])))

    def testOnModels(self):
        if IGNORE_TEST:
            return
        ffiles = os.listdir(MODEL_DIR)
        for ffile in ffiles:
            path = os.path.join(MODEL_DIR, ffile)
            try:
                rr = te.loada(path)
            except:
                continue
            antimony_str = rr.getAntimony()
            smat = rr.getFullStoichiometryMatrix()
            stoichiometry = Stoichiometry(antimony_str)
            self.assertTrue(np.all(stoichiometry.standard_mat == smat))
            self.assertTrue(all([x == y for x, y in zip(stoichiometry.species_names, smat.rownames)]))
            self.assertTrue(all([x == y for x, y in zip(stoichiometry.reaction_names, smat.colnames)]))

    def testWithFile(self):
        if IGNORE_TEST:
            return
        path = os.path.join(MODEL_DIR, "bestmodel_jJykOfGq0Kgy")
        rr = te.loada(path)
        antimony_str = rr.getAntimony()
        smat = rr.getFullStoichiometryMatrix()
        stoichiometry = Stoichiometry(antimony_str)
        self.assertTrue(np.all(stoichiometry.standard_mat == smat))
        self.assertTrue(all([x == y for x, y in zip(stoichiometry.species_names, smat.rownames)]))
        self.assertTrue(all([x == y for x, y in zip(stoichiometry.reaction_names, smat.colnames)]))


if __name__ == '__main__':
    unittest.main()